import os
import re
import json
import glob
import pandas as pd
import itertools


# Parse dataset metadata
###########################################
def metadata_parser(config):
    """Parse image metadata from a dataset.

    Parameters
    ----------
    config : plantcv.parallel.WorkflowConfig
        Workflow configuration object.

    Returns
    -------
    pandas.core.groupby.generic.DataFrameGroupBy
        Grouped dataframe of image metadata.
    """
    # Read the input dataset to a dictionary
    dataset = _read_dataset(config=config)

    # Convert the dataset metadata to a dataframe
    meta = _dataset2dataframe(dataset=dataset, config=config)

    # Split file paths into metadata terms 0:N-1
    meta = _parse_filepath(df=meta, config=config)

    # Apply user-supplied metadata filters
    meta, removed_df = _apply_metadata_filters(df=meta, config=config)

    # Apply user-supplied date range filters
    meta, removed_df = _apply_date_range_filter(df=meta, config=config, removed_df=removed_df)

    # Apply metadata grouping
    meta = _group_metadata(df=meta, config=config)

    return meta, removed_df
###########################################


# Parse dataset metadata
###########################################
def _read_dataset(config):
    """Read image datasets.

    Keyword arguments:
    config = plantcv.parallel.WorkflowConfig object

    Outputs:
    dataset = dataset metadata

    :param config: plantcv.parallel.WorkflowConfig
    :return dataset: dict
    """
    # Each dataset reader function outputs the dataset metadata in the same format
    # This makes the dataset compatible with the downstream steps
    # If you want to add a new dataset type, just add the logic here and create a new reader function

    # If the directory contains a metadata.json file it is a "phenodata" dataset
    if os.path.exists(os.path.join(config.input_dir, "metadata.json")):
        dataset = _read_phenodata(metadata_file=os.path.join(config.input_dir, "metadata.json"))
    # If the directory contains a SnapshotInfo.csv file it is a legacy "phenofront" dataset
    elif os.path.exists(os.path.join(config.input_dir, "SnapshotInfo.csv")):
        dataset = _read_phenofront(config=config, metadata_file=os.path.join(config.input_dir, "SnapshotInfo.csv"))
    else:
        dataset = _read_filenames(config=config)
    return dataset
###########################################


# Convert a dataset dictionary to a DataFrame
###########################################
def _dataset2dataframe(dataset, config):
    """Convert a dataset to a dataframe.

    Keyword arguments:
    dataset = dataset metadata dictionary
    config = plantcv.parallel.WorkflowConfig object

    Outputs:
    df = dataset metadata dataframe

    :param dataset: dict
    :param config: plantcv.parallel.WorkflowConfig
    :return df: pandas.core.frame.DataFrame
    """
    # Build a metadata dictionary of lists of metadata values
    metadata = {
        "filepath": [],
        "n_metadata_terms": []
    }
    # Populate metadata terms from the standard metadata vocabulary
    for term in config.metadata_terms:
        metadata[term] = []
    # Iterate over all image metadata and append metadata values to the dictionary
    for image in dataset["images"]:
        metadata["filepath"].append(os.path.join(config.input_dir, image))
        metadata["n_metadata_terms"].append(dataset["images"][image].get("n_metadata_terms"))
        for term in config.metadata_terms:
            metadata[term].append(dataset["images"][image].get(term))
    df = pd.DataFrame(data=metadata)
    utc = bool("Z" in config.timestampformat)
    df["timestamp"] = pd.to_datetime(df.timestamp, format=config.timestampformat, utc=utc)
    return df
###########################################


# Filter a metadata dataframe with user-supplied filters
###########################################
def _apply_metadata_filters(df, config):
    """Apply filters to metadata.
    Parameters
    ----------
    df = pandas.core.frame.Dataframe, metadata dataframe
    config = plantcv.parallel.WorkflowConfig object

    Returns:
    --------
    filtered_df = pandas.core.frame.Dataframe, filtered metadata dataframe
    removed_df  = pandas.core.frame.Dataframe, metadata dataframe of what was removed
    """
    # Convert all metadata filter values to a list type
    for term in config.metadata_filters:
        if not isinstance(config.metadata_filters[term], list):
            # Store term value as a list
            config.metadata_filters[term] = [config.metadata_filters[term]]
        # Convert filter values to strings
        config.metadata_filters[term] = list(map(str, config.metadata_filters[term]))
    # Create a metadata filter dataframe as the product of all combinations of values
    metadata_filter = pd.DataFrame(list(itertools.product(*config.metadata_filters.values())),
                                   columns=config.metadata_filters.keys(), dtype="object")
    # If there are no filters provide the metadata_filter dataframe will be empty and we can return the input datafram
    removed_df = pd.DataFrame()
    filtered_df = df
    if not metadata_filter.empty:
        filtered_df = df.merge(metadata_filter, how="inner")
        removed_df = _anti_join(df, filtered_df)
        removed_df["status"] = "Removed by config.metadata_filters"
        # if a row has None for all metadata then it was not able to be parsed due to variable length
        removed_df.loc[removed_df[list(config.metadata_filters.keys())].isnull().apply(all, axis=1),
                       'status'] = "Incorrect metadata length"
    if bool(config.metadata_regex):
        prev_df = filtered_df
        for key, value in config.metadata_regex.items():
            filtered_df = filtered_df[filtered_df[key].astype(str).str.contains(value, regex=True, na=False)]
        removed_df_2 = _anti_join(prev_df, filtered_df)
        removed_df_2["status"] = "Removed by config.metadata_regex"
        removed_df = pd.concat([removed_df, removed_df_2])
    return filtered_df, removed_df
###########################################


# Filter a metadata dataframe within a date range
###########################################
def _apply_date_range_filter(df, config, removed_df):
    """Filter metadata based on a date range.
    Parameters
    ----------
    df = pandas.core.frame.DataFrame, metadata dataframe
    config = plantcv.parallel.WorkflowConfig object
    removed_df = pandas.core.frame.DataFrame, dataframe of images removed up to this point

    Returns
    -------
    filtered_df = pandas.core.frame.DataFrame, filtered metadata dataframe
    removed_df = pandas.core.frame.DataFrame, dataframe of removed metadata
    """
    # Set whether the datetime code is in UTC or not
    utc = bool("Z" in config.timestampformat)

    # Include all by default
    after_start_date = pd.Series([True] * df.shape[0])
    before_end_date = pd.Series([True] * df.shape[0])

    # Make boolean vector for start and end date filtering if dates are not None
    if config.start_date is not None:
        after_start_date = df["timestamp"] >= pd.to_datetime(config.start_date, format=config.timestampformat, utc=utc)
    if config.end_date is not None:
        before_end_date = df["timestamp"] <= pd.to_datetime(config.end_date, format=config.timestampformat, utc=utc)

    # And of boolean vectors
    keep_dates = after_start_date & before_end_date
    # Keep rows with dates between start and end date
    filtered_df = df.loc[keep_dates]
    not_between_df = _anti_join(df, filtered_df)
    not_between_df["status"] = "Removed by config.start_date and config.end_date"
    removed_df = pd.concat([removed_df, not_between_df])

    return filtered_df, removed_df
###########################################


# Group metadata dataframe into sets
###########################################
def _group_metadata(df, config):
    """Group metadata.

    Keyword arguments:
    df = metadata dataframe
    config = plantcv.parallel.WorkflowConfig object

    Outputs:
    groups = grouped metadata

    :param df: pandas.core.frame.DataFrame
    :param config: plantcv.parallel.WorkflowConfig
    :return groups: pandas.core.groupby.generic.DataFrameGroupBy
    """
    groups = df.groupby(by=config.groupby)
    return groups
###########################################


# Initialize a dataset dictionary
###########################################
def _init_dataset():
    """Initialize a dataset."""
    dataset = {
        "dataset": {
            "experiment": ""
        },
        "environment": {},
        "images": {}
    }
    return dataset
###########################################


# Index positional filename metadata
###########################################
def _filename_metadata_index(config):
    """Index positional filename metadata.

    Parameters
    ----------
    config = plantcv.parallel.WorkflowConfig object

    Return
    ------
    metadata_index = dict, metadata terms and positions
    config = plantcv.parallel.WorkflowConfig object
    """
    # if filename_metadata is not specified then estimate it
    if not bool(config.filename_metadata):
        print("Warning: Creating config.filename_metadata based on file names.")
        config = _estimate_filename_metadata(config)

    # A dictionary of metadata terms and their index position in the filename metadata term list
    metadata_index = {}
    # Enumerate the terms listed in the user configuration
    for i, term in enumerate(config.filename_metadata):
        # Store the term and the listed order
        metadata_index[term] = i

    return metadata_index, config
###########################################


# Filename metadata parser
###########################################
def _parse_filename(filename, config, metadata_index):
    """Parse metadata from a filename.

    Parameters
    ----------
    filename = str, Filename to parse metadata from
    config = plantcv.parallel.WorkflowConfig object
    metadata_index = dict, dictionary of metadata terms and positions

    Returns
    -------
    img_meta = dict, dictionary of image metadata keys and valaues
    """
    # Image metadata
    img_meta = {}
    meta_list = []
    # Split the filename on delimiter if it is a single character
    if len(config.delimiter) == 1:
        meta_list = filename.split(config.delimiter)
    # Otherwise use a regular expression to parse metadata terms
    else:
        # Search using the supplied regular expression
        meta_list = re.search(config.delimiter, filename)
        # If there is a match convert the collected matches to a list
        if meta_list is not None:
            meta_list = list(meta_list.groups())
        # If thre is no match meta_list will be None, make an empty list
        else:
            meta_list = []
    # if all metadata terms start with "metadata_" then they are blank defaults and
    # we will include all pieces of the filepath.
    dummy_metadata = all(term.startswith("metadata_") for term in config.filename_metadata)
    if len(meta_list) == len(config.filename_metadata) or dummy_metadata:
        # For each of the type of metadata PlantCV keeps track of
        for i, term in enumerate(config.metadata_terms):
            # First store the default value for each term
            img_meta[term] = config.metadata_terms[term]["value"]
            # If the same metadata is found in the image filename, store the value
            if term in metadata_index:
                mi_term = metadata_index[term]
                img_meta[term] = None
                if i <= len(meta_list) - 1:
                    img_meta[term] = meta_list[mi_term]
    img_meta["n_metadata_terms"] = len(meta_list)
    return img_meta
###########################################


# Parses file paths into metadata columns
###########################################
def _parse_filepath(df, config):
    """Parse metadata from a filename.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe of image metadata.
    config : plantcv.parallel.WorkflowConfig
        PlantCV parallel configureation object.

    Returns
    -------
    pandas.DataFrame
        Dataframe with added filepath metadata columns.
    """
    # remove extraneous config.input_dir from file path
    paths_after_input = df["filepath"].map(lambda st: os.path.relpath(st, config.input_dir))
    path_metadata = []
    for i, fp in enumerate(paths_after_input):
        # for every file path, split it and add the elements to a list
        splits = fp.split(os.sep)
        path_metadata.append(splits[1:])
    # bind list into a dataframe
    path_metadata_df = pd.DataFrame(path_metadata)
    # rename columns to filepath1:N, basename
    path_metadata_df.columns = ["filepath"+str(i + 1) for i in range(len(path_metadata_df.columns))]
    if not path_metadata_df.empty:
        path_metadata_df.rename(columns={path_metadata_df.columns[-1]: "basename"}, inplace=True)
    # bind new columns onto existing metadata
    meta2 = df.reset_index(drop=True).join(path_metadata_df)
    return meta2
###########################################


# Reads "phenodata" datasets
###########################################
def _read_phenodata(metadata_file):
    """Read a phenodata metadata file.

    Keyword arguments:
    metadata_file = phenodata metadata.json file

    Outputs:
    dataset = dataset metadata

    :param metadata_file: str
    :return dataset: dict
    """
    # Reads metadata.json file from our LemnaTec database export tool
    # For phenodata datasets, all metadata is already parsed and stored in a JSON file
    with open(metadata_file, "r") as fp:
        dataset = json.load(fp)
        return dataset
###########################################


# Reads "phenofront" datasets
###########################################
def _read_phenofront(config, metadata_file):
    """Read phenofront metadata.

    Keyword arguments:
    config = plantcv.parallel.WorkflowConfig object
    metadata_file = phenofront SnapshotInfo.csv file

    Outputs:
    dataset = dataset metadata

    :param config: plantcv.parallel.WorkflowConfig
    :param metadata_file: str
    :return dataset: dict
    """
    # Create a dataset
    dataset = _init_dataset()
    # Index filename metadata based on user-supplied parsing parameters
    metadata_index, config = _filename_metadata_index(config=config)
    # if imgformat is all then set to png for legacy
    extension = config.imgformat
    if config.imgformat == "all":
        extension = "png"
    # Open the SnapshotInfo.csv file
    with open(metadata_file, 'r') as fp:
        # Read the first header line
        header = fp.readline()
        header = header.rstrip('\n')

        # Remove whitespace from the field names
        header = header.replace(" ", "")

        # Table column order
        cols = header.split(',')
        colnames = {}
        for i, col in enumerate(cols):
            colnames[col] = i

        # Read through the CSV file, each row is a snapshot
        for snapshot in fp:
            # Remove the line return
            snapshot = snapshot.rstrip('\n')
            # Split the snapshot metadata on commas
            snapshot_meta = snapshot.split(',')
            # Create the snapshot ID
            snapshot_id = f"snapshot{snapshot_meta[colnames['id']]}"
            # Store the snapshot metadata in the dataset dictionary
            dataset["environment"][snapshot_id] = {
                "barcode": snapshot_meta[colnames["plantbarcode"]],
                "cartag": snapshot_meta[colnames["cartag"]],
                "timestamp": snapshot_meta[colnames["timestamp"]],
                "weight_before": snapshot_meta[colnames["weightbefore"]],
                "weight_after": snapshot_meta[colnames["weightafter"]],
                "water_amount": snapshot_meta[colnames["wateramount"]],
                "completed": snapshot_meta[colnames["completed"]]
            }
            # Store the experiment name
            dataset["dataset"]["experiment"] = snapshot_meta[colnames["experiment"]]
            # Image names are stored in the tiles column
            img_list_str = snapshot_meta[colnames["tiles"]]
            # The tiles column can have a trailing semicolon
            img_list_str = img_list_str[:-1] if len(img_list_str) > 0 and img_list_str[-1] == ";" else img_list_str
            # Create a list of files by splitting on semicolon
            img_list = img_list_str.split(';')
            # Iterate over each image and store the metadata
            for img in img_list:
                # Parse camera label metaata
                img_meta = _parse_filename(filename=img, config=config, metadata_index=metadata_index)
                # Construct the filename
                filename = f"{img}.{extension}"
                # The dataset key is the dataset relative path to the image
                rel_path = os.path.join(snapshot_id, filename)
                # Store the parsed image metadata
                dataset["images"][rel_path] = img_meta
                # Update the metadata with metaata from SnapshotInfo.csv
                dataset["images"][rel_path].update({
                    "snapshot": snapshot_id,
                    "barcode": snapshot_meta[colnames["plantbarcode"]],
                    "cartag": snapshot_meta[colnames["cartag"]],
                    "timestamp": snapshot_meta[colnames["timestamp"]],
                    "camera_label": img
                })
    return dataset
###########################################


# Reads filename-based datasets
###########################################
def _read_filenames(config):
    """Read metadata from filenames.

    Keyword arguments:
    config = plantcv.parallel.WorkflowConfig object

    Outputs:
    dataset = dataset metadata

    :param config: plantcv.parallel.WorkflowConfig
    :return dataset: dict
    """
    # make imgformat a list if multiple
    extensions = config.imgformat
    if isinstance(config.imgformat, str):
        extensions = _replace_string_extension(config.imgformat)
    # Get a list of all files
    if config.include_all_subdirs is False:
        # If subdirectories are excluded, use glob to get a list of all image files
        fns = [f for ext in extensions for f in glob.glob(os.path.join(config.input_dir, "*[.]" + ext))]
    else:
        # If subdirectories are included, recursively walk through the path
        fns = []
        for root, _, files in os.walk(config.input_dir):
            for file in files:
                if file.lower().endswith(tuple(extensions)):
                    # Keep the files that end with the image extension
                    fns.append(os.path.join(root, file))
    # Create a dataset
    dataset = _init_dataset()
    # Name the experiment with the input directory
    dataset["dataset"]["experiment"] = config.input_dir
    # Index filename metadata based on user-supplied parsing parameters
    metadata_index, config = _filename_metadata_index(config=config)
    for filepath in fns:
        # Get the image dataset-relative path to use as the dataset key
        rel_path = os.path.relpath(filepath, start=config.input_dir)
        # Pull off the filename
        filename = os.path.basename(filepath)
        # Remove the extension
        metadata, _ = os.path.splitext(filename)
        # Store the image filename metadata
        dataset["images"][rel_path] = _parse_filename(filename=metadata, config=config, metadata_index=metadata_index)
    return dataset
###########################################


def _anti_join(df1, df2=None):
    """Anti join function for pandas dataframes
    Parameters
    ----------
    df1      = pandas.core.frame.Dataframe, dataframe of metadata
    df2      = pandas.core.frame.Dataframe, dataframe of filtered metadata

    Returns
    -------
    anti_joined = pandas.core.frame.Dataframe, metadata dataframe of what was removed
    """
    outer = df1.merge(df2, how='outer', indicator=True)
    anti_joined = outer[(outer['_merge'] == 'left_only')].drop('_merge', axis=1)
    return anti_joined
###########################################


# Reads filename-based datasets
###########################################

def _estimate_filename_metadata(config):
    """Estimate filename_metadata if it is missing
    Parameters
    ----------
    config = plantcv.parallel.WorkflowConfig object

    Returns
    -------
    config = plantcv.parallel.WorkflowConfig object with filename_metadata added
    """
    metadata_lengths = [1]
    imgformats = tuple(_replace_string_extension(config.imgformat))
    fns = []
    if config.include_all_subdirs is False:
        # If subdirectories are excluded, use glob to get a list of all image files
        for ext in imgformats:
            extfns = list(glob.glob(pathname=os.path.join(config.input_dir, f'*{ext}')))
            extfns = [os.path.basename(f) for f in extfns]
            fns.extend(extfns)
    else:
        fns = []
        for _, _, files in os.walk(config.input_dir):
            for file in files:
                if file.endswith(imgformats):
                    fns.append(file)
    # check length of metadata from all files, take the max, use those default terms.
    for file in fns:
        # get length of split filename
        metadata_lengths.append(len(file.split(config.delimiter)))
    config.filename_metadata = ["metadata_" + str(i) for i in range(max(metadata_lengths))]
    # if we had to make default metadata terms then add them to config.metadata_terms
    for term in config.filename_metadata:
        config.metadata_terms[term] = {
                "label": f"{term}",
                "datatype": "<class 'str'>",
                "value": "none"
            }

    return config
###########################################


def _replace_string_extension(imgformat):
    """Replace "all" with a list of file extensions.

    Parameters
    ----------
    imgformat : str
        The image format string, typically from a plantcv.parallel.WorkflowConfig object.

    Returns
    -------
    extensions : list of str
        A list of file extensions. If `imgformat` is "all", returns a list of common image file extensions;
        otherwise, returns a list containing only `imgformat`.
    """
    extensions = [imgformat]
    if imgformat == "all":
        extensions = ['bmp', 'dib', 'jpeg', 'jpg', 'jpe', 'jp2', 'png', 'ppm', 'pgm', 'ppm', 'sr', 'ras', 'tiff', 'tif']
    return extensions
###########################################



"""Analyzes the temperature values of objects in an image."""
import os
import numpy as np
from plantcv.plantcv import params, outputs
from plantcv.plantcv._debug import _debug
from plantcv.plantcv.visualize import histogram
from plantcv.plantcv._helpers import _iterate_analysis


def thermal(thermal_img, labeled_mask, n_labels=1, bins=100, label=None):
    """Analyzes the temperature values of objects in an image.

    Inputs:
    thermal_img  = Thermal image data.
    labeled_mask = Labeled mask of objects (32-bit).
    n_labels     = Total number expected individual objects (default = 1).
    bins         = Number of histogram bins (default = 100)
    label        = optional label parameter, modifies the variable name of observations recorded (default = "default").

    Returns:
    analysis_image = Thermal histogram plot

    :param thermal_img: numpy.ndarray
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param bins: int
    :param label: str
    :return analysis_image: altair.vegalite.v5.api.FacetChart
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    _ = _iterate_analysis(img=thermal_img, labeled_mask=labeled_mask, n_labels=n_labels, label=label,
                          function=_analyze_thermal,  **{"bins": bins})
    temp_chart = outputs.plot_dists(variable="thermal_frequencies")
    _debug(visual=temp_chart, filename=os.path.join(params.debug_outdir, str(params.device) + '_temperature_hist.png'))
    return temp_chart


def _analyze_thermal(img, mask, bins=100, label=None):
    """Extract the temperature values of an object in an image.

    Inputs:
    thermal_img  = numpy array of thermal values
    mask         = Binary mask made from selected contours
    bins         = Number of histogram bins (default = 100)
    label        = optional label parameter, modifies the variable name of observations recorded

    Returns:
    thermal_img  = output image

    :param thermal_array: numpy.ndarray
    :param mask: numpy.ndarray
    :param bins: int
    :param label: str
    :return thermal_img: numpy.ndarray
    """
    # Store debug mode
    debug = params.debug

    # apply plant shaped mask to image and calculate statistics based on the masked image
    masked_thermal = img[np.where(mask > 0)]
    maxtemp = np.amax(masked_thermal)
    mintemp = np.amin(masked_thermal)
    avgtemp = np.average(masked_thermal)
    mediantemp = np.median(masked_thermal)

    # call the histogram function
    params.debug = None
    _, hist_data = histogram(img, mask=mask, bins=bins, hist_data=True)
    bin_labels, hist_percent = hist_data['pixel intensity'].tolist(), hist_data['proportion of pixels (%)'].tolist()

    # Store data into outputs class
    outputs.add_observation(sample=label, variable='max_temp', trait='maximum temperature',
                            method='plantcv.plantcv.analyze.thermal', scale='degrees', datatype=float,
                            value=maxtemp, label='degrees')
    outputs.add_observation(sample=label, variable='min_temp', trait='minimum temperature',
                            method='plantcv.plantcv.analyze.thermal', scale='degrees', datatype=float,
                            value=mintemp, label='degrees')
    outputs.add_observation(sample=label, variable='mean_temp', trait='mean temperature',
                            method='plantcv.plantcv.analyze.thermal', scale='degrees', datatype=float,
                            value=avgtemp, label='degrees')
    outputs.add_observation(sample=label, variable='median_temp', trait='median temperature',
                            method='plantcv.plantcv.analyze.thermal', scale='degrees', datatype=float,
                            value=mediantemp, label='degrees')
    outputs.add_observation(sample=label, variable='thermal_frequencies', trait='thermal frequencies',
                            method='plantcv.plantcv.analyze.thermal', scale='frequency', datatype=list,
                            value=hist_percent, label=bin_labels)
    # Restore user debug setting
    params.debug = debug

    return img



"""Watershed segmentation function."""
# This function is based on code contributed by Suxing Liu, Arkansas State University.
# For more information see https://github.com/lsx1980/Leaf_count

import cv2
import os
import numpy as np
from scipy import ndimage as ndi
from skimage.feature import peak_local_max
from skimage.segmentation import watershed
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import color_palette
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv._helpers import _cv2_findcontours


def watershed_segmentation(rgb_img, mask, distance=10, label=None):
    """
    Uses the watershed algorithm to detect boundary of objects. Needs a marker file which specifies area which is
    object (white), background (grey), unknown area (black).

    Inputs:
    rgb_img             = image to perform watershed on needs to be 3D (i.e. np.shape = x,y,z not np.shape = x,y)
    mask                = binary image, single channel, object in white and background black
    distance            = min_distance of local maximum
    label               = Optional label parameter, modifies the variable name of
                          observations recorded (default = pcv.params.sample_label).

    Returns:
    labels              = labeled mask with segmentation results

    :param rgb_img: numpy.ndarray
    :param mask: numpy.ndarray
    :param distance: int
    :param label: str
    :return labels: numpy.ndarray
    """
    # Store color sequence mode and set to random for watershed_img debug
    color_sequence = params.color_sequence
    params.color_sequence = "random"

    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    dist_transform = cv2.distanceTransformWithLabels(mask, cv2.DIST_L2, maskSize=0)[0]

    local_max_coordinates = peak_local_max(dist_transform, min_distance=distance, labels=mask)
    local_max = np.zeros_like(dist_transform, dtype=bool)
    local_max[tuple(local_max_coordinates.T)] = True

    markers = ndi.label(local_max, structure=np.ones((3, 3)))[0]
    dist_transform1 = -dist_transform
    labels = watershed(dist_transform1, markers, mask=mask)
    estimated_object_count = len(np.unique(markers)) - 1

    # Plot image
    plt_img = np.copy(rgb_img)
    rand_color = color_palette(len(np.unique(labels)))
    for i in np.unique(labels):
        # Skip black background i=0
        if i > 0:
            # Find contours
            submask = np.where(labels == i, 255, 0).astype(np.uint8)
            cnt, _ = _cv2_findcontours(bin_img=submask)
            cv2.drawContours(plt_img, cnt, -1, rand_color[i], params.line_thickness)

    _debug(visual=plt_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + '_watershed_labeled_img.png'),
           cmap='gray')

    # Reset color sequence mode
    params.color_sequence = color_sequence

    outputs.add_observation(sample=label, variable='estimated_object_count', trait='estimated object count',
                            method='plantcv.plantcv.watershed', scale='none', datatype=int,
                            value=estimated_object_count, label='none')

    return labels



"""Function to scan for pseudolandmarks along the y-axis."""
import cv2
import numpy as np
from plantcv.plantcv._helpers import _cv2_findcontours, _object_composition
from plantcv.plantcv.homology.x_axis_pseudolandmark import _small_img_pseudolandmarks, _pseudolandmarks
from plantcv.plantcv import params


def y_axis_pseudolandmarks(img, mask, label=None):
    """Divide up object contour into 19 equidistant segments and generate landmarks for each.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting
    mask : numpy.ndarray
        Binary mask of the object of interest
    label : str or None, optional
        Optional label parameter, modifies the variable name of
        observations recorded, (default = pcv.params.sample_label)

    Returns
    -------
    left : list
        Left landmark points (depending on axis)
    right : list
        Right landmark points (depending on axis)
    center_h : list
        Horizontal landmark points in middle portion (depending on axis)
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    # Find contours
    cnt, cnt_str = _cv2_findcontours(bin_img=mask)
    # Consolidate contours
    obj = _object_composition(contours=cnt, hierarchy=cnt_str)
    # Empty pot type scenario
    if not np.any(obj):
        return ('NA', 'NA'), ('NA', 'NA'), ('NA', 'NA')
    # Bounding rectangle
    _, _, _, height = cv2.boundingRect(obj)
    extent = height
    # If height is greater than 21 pixels make 20 increments (5% intervals)
    if extent >= 21:
        left, right, center_h = _pseudolandmarks(img, obj, mask, label, 1)
        # If the width of the object is less than 20 pixels just make the object a 20 pixel rectangle
    elif extent < 21:
        left, right, center_h = _small_img_pseudolandmarks(img, obj, mask, label, 1)

    return left, right, center_h



# Remove small bright spots

import os
import numpy as np
from skimage import morphology
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rect_filter, _rect_replace
from plantcv.plantcv import fatal_error


def opening(gray_img, kernel=None, roi=None):
    """
    Wrapper for scikit-image opening functions. Opening can remove small bright spots (i.e. salt).

    Inputs:
    gray_img = input image (grayscale or binary)
    kernel   = optional neighborhood, expressed as an array of 1s and 0s. If None, use cross-shaped structuring element.
    roi      = optional rectangular ROI to open within

    :param gray_img: ndarray
    :param kernel: ndarray
    :param roi: plantcv.plantcv.Objects
    :return filtered_img: ndarray
    """
    # Make sure the image is binary/grayscale
    if len(np.shape(gray_img)) != 2:
        fatal_error("Input image must be grayscale or binary")

    # If image is binary use the faster method
    if len(np.unique(gray_img)) == 2:
        bool_img = gray_img.astype(bool)
        sub_img = _rect_filter(bool_img, roi, function=morphology.binary_opening,
                               **{"footprint": kernel})
        filtered_img = sub_img.astype(np.uint8) * 255
        replaced_img = _rect_replace(bool_img.astype(np.uint8) * 255, filtered_img, roi)
    # Otherwise use method appropriate for grayscale images
    else:
        filtered_img = _rect_filter(gray_img, roi=roi, function=morphology.opening,
                                    **{"footprint": kernel})
        replaced_img = _rect_replace(gray_img, filtered_img, roi)

    _debug(visual=replaced_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + '_opening.png'),
           cmap='gray')

    return replaced_img



"""Plot image to screen."""
import cv2
import numpy
import matplotlib
from xarray.core.dataarray import DataArray
from plantcv.plantcv import fatal_error, params
from plantcv.plantcv.classes import PSII_data
from plantcv.plantcv._show_dataarray import _show_dataarray
from matplotlib import pyplot as plt
from altair.vegalite.v5.api import FacetChart, LayerChart, Chart


def plot_image(img, cmap=None, **kwargs):
    """
    Plot an image to the screen.

    :param img: numpy.ndarray, ggplot, xarray.core.dataarray.DataArray
    :param cmap: str
    :param kwargs: key-value arguments to xarray.plot method
    :return:
    """
    dimensions = numpy.shape(img)

    if isinstance(img, numpy.ndarray):
        matplotlib.rcParams['figure.dpi'] = params.dpi
        # If the image is color then OpenCV stores it as BGR, we plot it as RGB
        if len(dimensions) == 3:
            plt.figure()
            plt.imshow(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
            plt.show()

        elif cmap is None and len(dimensions) == 2:
            plt.figure()
            plt.imshow(img, cmap="gray")
            plt.show()

        elif cmap is not None and len(dimensions) == 2:
            plt.figure()
            plt.imshow(img, cmap=cmap)
            plt.show()

    elif isinstance(img, matplotlib.figure.Figure):
        fatal_error(
            "Error, matplotlib Figure not supported. Instead try running without plot_image.")

    elif isinstance(img, DataArray):
        _show_dataarray(img, **kwargs)

    # Altair FacetChart
    elif isinstance(img, (FacetChart, LayerChart, Chart)):
        img.display()

    elif isinstance(img, PSII_data):
        fatal_error("You need to plot an underlying DataArray.")

    else:
        fatal_error(f"Plotting {type(img)} is not supported.")



"""Analyzes an object and outputs numeric properties."""
import cv2
import numpy as np
import os
from plantcv.plantcv import params, outputs, fatal_error, apply_mask, warn, deprecation_warning
from plantcv.plantcv.threshold import binary as binary_threshold
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _object_composition, _roi_filter, _rgb2hsv


def report_size_marker_area(img, roi, marker='define', objcolor='dark', thresh_channel=None,
                            thresh=None, label=None):
    """
    Detects a size marker in a specified region and reports its size and eccentricity

    Inputs:
    img             = An RGB or grayscale image to plot the marker object on
    roi             = A region of interest (e.g. output from pcv.roi.rectangle or other methods)
    marker          = 'define' or 'detect'. If define it means you set an area, if detect it means you want to
                      detect within an area
    objcolor        = Object color is 'dark' or 'light' (is the marker darker or lighter than the background)
    thresh_channel  = 'h', 's', or 'v' for hue, saturation or value
    thresh          = Binary threshold value (integer)
    label           = Optional label parameter, modifies the variable name of
                      observations recorded (default = pcv.params.sample_label).

    Returns:
    analysis_images = List of output images

    :param img: numpy.ndarray
    :param roi: plantcv.plantcv.classes.Objects
    :param marker: str
    :param objcolor: str
    :param thresh_channel: str
    :param thresh: int
    :param label: str
    :return: analysis_images: list
    """
    deprecation_warning(
        "the 'label' parameter is no longer utilized, since size marker is now metadata. "
        "It will be removed in PlantCV 5.0."
        )

    # Store debug
    debug = params.debug
    params.debug = None
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    params.device += 1
    # Make a copy of the reference image
    ref_img = np.copy(img)
    # If the reference image is grayscale convert it to color
    if len(np.shape(ref_img)) == 2:
        ref_img = cv2.cvtColor(ref_img, cv2.COLOR_GRAY2BGR)

    # Marker components
    # If the marker type is "defined" then the marker_mask and marker_contours are equal to the input ROI
    # Initialize a binary image
    roi_mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
    # Draw the filled ROI on the mask
    cv2.drawContours(roi_mask, roi.contours[0], -1, (255), -1)
    # Marker mask
    marker_mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)

    # If the marker type is "detect" then we will use the ROI to isolate marker contours from the input image
    if marker.upper() == 'DETECT':
        # We need to convert the input image into an one of the HSV channels and then threshold it
        if thresh_channel is not None and thresh is not None:
            # Mask the input image
            masked = apply_mask(img=ref_img, mask=roi_mask, mask_color="black")
            # Convert the masked image to hue, saturation, or value
            marker_hsv = _rgb2hsv(rgb_img=masked, channel=thresh_channel)
            # Threshold the HSV image
            marker_bin = binary_threshold(gray_img=marker_hsv, threshold=thresh, object_type=objcolor)
            # Identify contours in the masked image
            contours, hierarchy = _cv2_findcontours(bin_img=marker_bin)

            # Filter marker contours using the input ROI
            kept_contours, kept_hierarchy, _ = _roi_filter(img=marker_bin, roi=roi, obj=contours,
                                                           hierarchy=hierarchy, roi_type="partial")
            # If there are more than one contour detected, combine them into one
            marker_contour = _object_composition(contours=kept_contours, hierarchy=kept_hierarchy)
            cv2.drawContours(marker_mask, kept_contours, -1, (255), -1, hierarchy=kept_hierarchy)
        else:
            # Reset debug mode
            params.debug = debug
            fatal_error('thresh_channel and thresh must be defined in detect mode')
    elif marker.upper() == "DEFINE":
        marker_mask = roi_mask
    else:
        # Reset debug mode
        params.debug = debug
        fatal_error(f"marker must be either 'define' or 'detect' but {marker} was provided.")

    # Calculate the moments of the defined marker region
    m = cv2.moments(marker_mask, binaryImage=True)
    cnt, h = _cv2_findcontours(bin_img=marker_mask)
    marker_contour = _object_composition(contours=cnt, hierarchy=h)

    # Calculate the marker area
    marker_area = m['m00']

    if len(marker_contour) > 5:
        # Fit a bounding ellipse to the marker
        _, axes, _ = cv2.fitEllipse(marker_contour)
        major_axis = np.argmax(axes)
        minor_axis = 1 - major_axis
        major_axis_length = axes[major_axis]
        minor_axis_length = axes[minor_axis]
        # Calculate the bounding ellipse eccentricity
        eccentricity = np.sqrt(1 - (axes[minor_axis] / axes[major_axis]) ** 2)

        cv2.drawContours(ref_img, marker_contour, -1, (255, 0, 0), 5)
        analysis_image = ref_img

        # Reset debug mode
        params.debug = debug

        _debug(visual=ref_img,
               filename=os.path.join(params.debug_outdir, str(params.device) + '_marker_shape.png'))

        # Store size marker values as metadata
        outputs.add_metadata(term="marker_area", datatype=int, value=marker_area)
        outputs.add_metadata(term="marker_ellipse_major_axis", datatype=int, value=major_axis_length)
        outputs.add_metadata(term="marker_ellipse_minor_axis", datatype=int, value=minor_axis_length)
        outputs.add_metadata(term="marker_ellipse_eccentricity", datatype=float, value=eccentricity)

        return analysis_image

    # Store size marker values as metadata
    outputs.add_metadata(term="marker_area", datatype=str, value='none')
    outputs.add_metadata(term="marker_ellipse_major_axis", datatype=str, value='none')
    outputs.add_metadata(term="marker_ellipse_minor_axis", datatype=str, value='none')
    outputs.add_metadata(term="marker_ellipse_eccentricity", datatype=str, value='none')

    warn("Size marker is not detectable.")

    return None



# Erosion filter

import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _erode
from plantcv.plantcv import params


def erode(gray_img, ksize, i, roi=None):
    """Perform morphological 'erosion' filtering.

    Keeps pixel in center of the kernel if conditions set in kernel are
       true, otherwise removes pixel.

    Parameters
    ----------
    gray_img : numpy.ndarray
             Grayscale (usually binary) image data
    ksize : int
             Kernel size (int). A ksize x ksize kernel will be built. Must be greater than 1 to have an effect.
    i : int
             interations, i.e. number of consecutive filtering passes
    roi : plantcv.plantcv.Objects
             Optional rectangular ROI to erode within

    Returns
    -------
    numpy.ndarray
         Eroded result image

    Raises
    ------
    ValueError
        If ksize is less than or equal to 1.
    """
    er_img = _erode(gray_img, ksize, i, roi=roi)

    _debug(er_img,
           filename=os.path.join(params.debug_outdir,
                                 str(params.device) + '_er_image' + str(ksize) + '_itr_' + str(i) + '.png'),
           cmap='gray')

    return er_img



# ROI functions

import os
import cv2
import numpy as np
import pandas as pd
from sklearn.mixture import GaussianMixture
from skimage.measure import label
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import color_palette
from plantcv.plantcv._helpers import _cv2_findcontours
from plantcv.plantcv._helpers import _roi_filter
from plantcv.plantcv._helpers import _hough_circle
from plantcv.plantcv._helpers import _rect_filter, _rect_replace
from plantcv.plantcv import fatal_error, warn, params, Objects


# Create an ROI from a binary mask
def from_binary_image(img, bin_img):
    """Create an ROI from a binary image

    Inputs:
    img           = An RGB or grayscale image to plot the ROI on.
    bin_img       = Binary image to extract an ROI contour from.

    Outputs:
    roi

    :param img: numpy.ndarray
    :param bin_img: numpy.ndarray
    :return roi: plantcv.plantcv.classes.Objects
    """
    # Make sure the input bin_img is binary
    if len(np.unique(bin_img)) != 2:
        fatal_error("Input image is not binary!")
    # Use the binary image to create an ROI contour
    roi_contour, roi_hierarchy = _cv2_findcontours(bin_img=bin_img)
    roi = Objects(contours=[roi_contour], hierarchy=[roi_hierarchy])
    # Draw the ROI if requested
    _draw_roi(img=img, roi_contour=roi)

    return roi


# Create a rectangular ROI
def rectangle(img, x, y, h, w):
    """Create a rectangular ROI.

    Inputs:
    img           = An RGB or grayscale image to plot the ROI on in debug mode.
    x             = The x-coordinate of the upper left corner of the rectangle.
    y             = The y-coordinate of the upper left corner of the rectangle.
    h             = The height of the rectangle.
    w             = The width of the rectangle.

    Outputs:
    roi

    :param img: numpy.ndarray
    :param x: int
    :param y: int
    :param h: int
    :param w: int
    :return roi: plantcv.plantcv.classes.Objects
    """
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]

    # Create the rectangle contour vertices
    pt1 = [x, y]
    pt2 = [x, y + h - 1]
    pt3 = [x + w - 1, y + h - 1]
    pt4 = [x + w - 1, y]

    # Create the ROI contour
    roi_contour = [np.array([[pt1], [pt2], [pt3], [pt4]], dtype=np.int32)]
    roi_hierarchy = np.array([[[-1, -1, -1, -1]]], dtype=np.int32)
    roi = Objects(contours=[roi_contour], hierarchy=[roi_hierarchy])

    # Draw the ROI if requested
    _draw_roi(img=img, roi_contour=roi)

    # Check whether the ROI is correctly bounded inside the image
    if x < 0 or y < 0 or x + w > width or y + h > height:
        fatal_error("The ROI extends outside of the image!")

    return roi


# Create a circular ROI
def circle(img, x, y, r):
    """Create a circular ROI.

    Inputs:
    img           = An RGB or grayscale image to plot the ROI on in debug mode.
    x             = The x-coordinate of the center of the circle.
    y             = The y-coordinate of the center of the circle.
    r             = The radius of the circle.

    Outputs:
    roi           = A dataclass with the roi object and hierarchy.

    :param img: numpy.ndarray
    :param x: int
    :param y: int
    :param r: int
    :return roi: plantcv.plantcv.classes.Objects
    """
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]

    # Initialize a binary image of the circle
    bin_img = np.zeros((height, width), dtype=np.uint8)
    # Draw the circle on the binary image
    cv2.circle(bin_img, (x, y), r, 255, -1)

    # Use the binary image to create an ROI contour
    roi_contour, roi_hierarchy = _cv2_findcontours(bin_img=bin_img)
    roi = Objects(contours=[roi_contour], hierarchy=[roi_hierarchy])

    # Draw the ROI if requested
    _draw_roi(img=img, roi_contour=roi)

    # Check whether the ROI is correctly bounded inside the image
    if x - r < 0 or x + r > width or y - r < 0 or y + r > height:
        fatal_error("The ROI extends outside of the image!")

    return roi


# Create an elliptical ROI
def ellipse(img, x, y, r1, r2, angle):
    """Create an elliptical ROI.

    Inputs:
    img           = An RGB or grayscale image to plot the ROI on in debug mode.
    x             = The x-coordinate of the center of the ellipse.
    y             = The y-coordinate of the center of the ellipse.
    r1            = The radius of the minor axis.
    r2            = The radius of the major axis.
    angle         = The angle of rotation in degrees of the major axis.

    Outputs:
    roi           = a dataclass with the roi object and hierarchy

    :param img: numpy.ndarray
    :param x: int
    :param y: int
    :param r1: int
    :param r2: int
    :param angle: double
    :return roi: plantcv.plantcv.classes.Objects
    """
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]

    # Initialize a binary image of the ellipse
    bin_img = np.zeros((height, width), dtype=np.uint8)
    # Draw the ellipse on the binary image
    cv2.ellipse(bin_img, (x, y), (r1, r2), angle, 0, 360, 255, -1)

    # Use the binary image to create an ROI contour
    roi_contour, roi_hierarchy = _cv2_findcontours(bin_img=bin_img)
    roi = Objects(contours=[roi_contour], hierarchy=[roi_hierarchy])

    # Draw the ROI if requested
    _draw_roi(img=img, roi_contour=roi)

    # Checks ellipse goes outside the image by checking row and column sum of edges
    if (np.sum(bin_img[0, :]) + np.sum(bin_img[-1, :]) + np.sum(bin_img[:, 0]) + np.sum(bin_img[:, -1]) > 0) or \
            len(roi_contour) == 0:
        fatal_error("The ROI extends outside of the image, or ROI is not on the image!")

    return roi


# Draw the ROI on a reference image
def _draw_roi(img, roi_contour):
    """Draw an ROI

    Parameters
    ----------
    img : numpy.ndarray
        An RGB or grayscale image to plot the ROI on in debug mode.
    roi_contour : list
        A list of contours and hierarchies for the ROI.
    """
    # Make a copy of the reference image
    ref_img = np.copy(img)
    # If the reference image is grayscale convert it to color
    if len(np.shape(ref_img)) == 2:
        ref_img = cv2.cvtColor(ref_img, cv2.COLOR_GRAY2BGR)
    # Collect coordinates for debug numbering
    rand_color = color_palette(num=len(roi_contour.contours),
                               saved=False) if len(roi_contour.contours) > 1 else [params.line_color]
    label_coords = []
    for i, cnt in enumerate(roi_contour):
        M = cv2.moments(cnt.contours[0][0])
        if M['m00'] != 0:
            cxy = [int(M['m10'] / M['m00']), int(M['m01'] / M['m00'])]
            label_coords.append(cxy)
            # Add number labels to debug
            cv2.putText(img=ref_img, text=f"{i+1}", org=(label_coords[i]),
                        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale=params.text_size, color=rand_color[i],
                        thickness=params.text_thickness)
        # Draw ROI outline regardless
        cv2.drawContours(ref_img, cnt.contours[0], -1, rand_color[i], params.line_thickness)

    _debug(visual=ref_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_roi.png"))


def _calculate_grid(mask, nrows, ncols):
    """Calculate the grid layout of ROIs

    Parameters
    ----------
    mask : numpy.ndarray
        A binary mask
    nrows : int
        Number of rows in ROI layout
    ncols : int
        Number of columns in ROI layout

    Returns
    -------
    tuple, tuple
        Two-element tuple of the center of the top left object (x,y) and Two-element tuple of the horizontal and vertical
        spacing between ROIs, (x,y)
    """
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[-2:]
    centers = []
    for c in contours:
        m = cv2.moments(c)
        if m['m00'] != 0:
            cmx, cmy = (float(m['m10'] / m['m00']), float(m['m01'] / m['m00']))
            centers.append((cmx, cmy))
    # cluster by x and y coordinates to get grid layout
    centers_x = np.array(np.array([i[0] for i in centers]).reshape(-1, 1))
    centers_y = np.array(np.array([i[1] for i in centers]).reshape(-1, 1))
    gm_x = GaussianMixture(n_components=ncols, random_state=0).fit(centers_x)
    gm_y = GaussianMixture(n_components=nrows, random_state=0).fit(centers_y)
    clusters_x = np.sort(gm_x.means_[:, 0])
    clusters_y = np.sort(gm_y.means_[:, 0])
    spacing_x = (clusters_x[ncols-1] - clusters_x[0])/(ncols-1) if ncols > 1 else 0
    spacing_y = (clusters_y[nrows-1] - clusters_y[0])/(nrows-1) if nrows > 1 else 0
    spacing = (round(spacing_x), round(spacing_y))
    coord = (round(clusters_x[0]), round(clusters_y[0]))
    return coord, spacing


def _adjust_radius_coord(height, width, coord, radius):
    """Adjust the radius of the ROI

    Parameters
    ----------
    height : int
        Height of the image
    width : int
        Width of the image
    coord : list
        List of tuples identifying the center of each roi [(x1,y1),(x2,y2)]
    radius : int
        Radius of the ROI

    Returns
    -------
    int
        Adjusted radius
    """
    x = [i[0] for i in coord]
    y = [i[1] for i in coord]
    return _adjust_radius_max_min(height, width, radius, max(x), min(x), max(y), min(y))


def _adjust_radius_grid(height, width, coord, radius, spacing, nrows, ncols):
    """Adjust the radius of the ROI

    Parameters
    ----------
    height : int
        Height of the image
    width : int
        Width of the image
    coord : tuple
        Two-element tuple of the center of the top left object (x,y)
    radius : int
        Radius of the ROI
    spacing : tuple
        Two-element tuple of the horizontal and vertical spacing between ROIs, (x,y)
    nrows : int
        Number of rows in ROI layout
    ncols : int
        Number of columns in ROI layout

    Returns
    -------
    int
        Adjusted radius
    """
    xmax = coord[0] + (ncols-1)*spacing[0]
    xmin = coord[0]
    ymax = coord[1] + (nrows-1)*spacing[1]
    ymin = coord[1]
    return _adjust_radius_max_min(height, width, radius, xmax, xmin, ymax, ymin)


def _adjust_radius_max_min(height, width, radius, xmax, xmin, ymax, ymin):
    """Adjust the radius of the ROI

    Parameters
    ----------
    height : int
        Height of the image
    width : int
        Width of the image
    radius : int
        Radius of the ROI
    xmax : int
        Maximum x coordinate of the ROI
    xmin : int
        Minimum x coordinate of the ROI
    ymax : int
        Maximum y coordinate of the ROI
    ymin : int
        Minimum y coordinate of the ROI

    Returns
    -------
    int
        Adjusted radius
    """
    if ((xmin < 0) or (xmax > width) or (ymin < 0) or (ymax > height)):
        fatal_error("An ROI extends outside of the image!")
    distances_to_edge = [xmin, width-xmax, ymin, height-ymax]
    min_distance = min(distances_to_edge)
    if min_distance < radius:
        warn('Shrinking radius to make ROIs fit in the image')
        radius = min_distance - 1
    return radius


def _rois_from_coordinates(img, coord=None, radius=None):
    """Create multiple circular ROIs on a single image from a list of coordinates

    Parameters
    ----------
    img : numpy.ndarray
        Input image data
    coord : list, optional
        List of tuples identifying the center of each roi [(x1,y1),(x2,y2)], by default None
    radius : int, optional
        A single radius for all ROIs, by default None

    Returns
    -------
    plantcv.plantcv.classes.Objects, numpy.ndarray
        A dataclass with roi objects and hierarchies, A binary image with overlapping ROIs
    """
    if radius is None:
        fatal_error("Specify a radius if creating rois from a list of coordinates")
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]
    radius = _adjust_radius_coord(height, width, coord, radius)
    overlap_img = np.zeros((height, width), dtype=np.uint8)
    roi_objects = Objects()
    for i in range(0, len(coord)):
        # Initialize a binary image for each circle
        bin_img = np.zeros((height, width), dtype=np.uint8)
        y = int(coord[i][1])
        x = int(coord[i][0])
        # Draw the circle on the binary image
        # Keep track of each roi individually to check overlapping
        circle_img = cv2.circle(bin_img, (x, y), radius, 255, -1)
        overlap_img = cv2.circle(overlap_img, (x, y), radius, 255, -1)
        # Make a list of contours and hierarchies
        rc, rh = cv2.findContours(circle_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[-2:]
        roi_objects.append(rc, rh)
    return roi_objects, overlap_img


def _rect_rois_from_coordinates(img, h, w, coord=None):
    """Create multiple circular ROIs on a single image from a list of coordinates

    Parameters
    ----------
    img : numpy.ndarray
        Input image data
    coord : list, optional
        List of tuples identifying the center of each roi [(x1,y1),(x2,y2)], by default None
    h : int
        Height of rectangular ROIs
    w : int
        Width of rectangular ROIs

    Returns
    -------
    plantcv.plantcv.classes.Objects, numpy.ndarray
        A dataclass with roi objects and hierarchies, A binary image with overlapping ROIs
    """
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]
    overlap_img = np.zeros((height, width), dtype=np.uint8)
    roi_objects = Objects()
    for i in range(0, len(coord)):
        y = int(coord[i][1])
        x = int(coord[i][0])
        # find corners of rectangle
        pt1 = [x, y]
        pt2 = [x, y + h - 1]
        pt3 = [x + w - 1, y + h - 1]
        pt4 = [x + w - 1, y]
        # Make a list of contours and hierarchies
        rc = [np.array([[pt1], [pt2], [pt3], [pt4]], dtype=np.int32)]
        rh = np.array([[[-1, -1, -1, -1]]], dtype=np.int32)
        # Keep track of each roi individually to check overlapping
        overlap_img = cv2.drawContours(overlap_img, rc, -1, 255, -1)
        roi_objects.append(rc, rh)
    return roi_objects, overlap_img


def _grid_roi(img, nrows, ncols, coord=None, radius=None, spacing=None):
    """Create a grid of ROIs

    Parameters
    ----------
    img : numpy.ndarray
        Input image data
    nrows : int
        Number of rows in ROI layout
    ncols : int
        Number of columns in ROI layout
    coord : tuple, optional
        Two-element tuple of the center of the top left object (x,y), by default None
    radius : int, optional
        Optional parameter to specify the radius of the circular rois, by default None
    spacing : tuple, optional
        Two-element tuple of the horizontal and vertical spacing between ROIs, (x,y), by default None

    Returns
    -------
    plantcv.plantcv.classes.Objects, numpy.ndarray
        A dataclass with roi objects and hierarchies, A binary image with overlapping ROIs
    """
    if radius is None:
        RADIUS_RATIO = 0.325
        if spacing[0] == 0:
            radius = round(RADIUS_RATIO*spacing[1])
        elif spacing[1] == 0:
            radius = round(RADIUS_RATIO*spacing[0])
        else:
            radius = round(RADIUS_RATIO*(spacing[0]+spacing[1])/2)
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]
    radius = _adjust_radius_grid(height, width, coord, radius, spacing, nrows, ncols)
    overlap_img = np.zeros((height, width), dtype=np.uint8)
    roi_objects = Objects()
    # Loop over each row
    for i in range(0, nrows):
        # The upper left corner is the y starting coordinate + the ROI offset * the vertical spacing
        y = coord[1] + i * spacing[1]
        # Loop over each column
        for j in range(0, ncols):
            # Initialize a binary image for each circle
            bin_img = np.zeros((height, width), dtype=np.uint8)
            # The upper left corner is the x starting coordinate + the ROI offset * the
            # horizontal spacing between chips
            x = coord[0] + j * spacing[0]
            # Draw the circle on the binary images
            # Keep track of each roi individually to check overlapping
            circle_img = cv2.circle(bin_img, (x, y), radius, 255, -1)
            # Draw the circle on the overall mask
            overlap_img = cv2.circle(overlap_img, (x, y), radius, 255, -1)
            # Make a list of contours and hierarchies
            rc, rh = cv2.findContours(circle_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[-2:]
            roi_objects.append(rc, rh)
    return roi_objects, overlap_img


def _grid_roi_rect(img, nrows, ncols, h, w, coord=None, spacing=None):
    """Create a grid of rectangular ROIs

    Parameters
    ----------
    img : numpy.ndarray
        Input image data
    nrows : int
        Number of rows in ROI layout
    ncols : int
        Number of columns in ROI layout
    coord : tuple, optional
        Two-element tuple of the center of the top left object (x,y), by default None
    h : int
        Height of each rectangle
    w : int
        Width of each rectangle
    spacing : tuple, optional
        Two-element tuple of the horizontal and vertical spacing between ROIs, (x,y), by default None

    Returns
    -------
    plantcv.plantcv.classes.Objects, numpy.ndarray
        A dataclass with roi objects and hierarchies, A binary image with overlapping ROIs
    """
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]
    overlap_img = np.zeros((height, width), dtype=np.uint8)
    roi_objects = Objects()
    # Loop over each row
    for i in range(0, nrows):
        # The upper left corner is the y starting coordinate + the ROI offset * the vertical spacing
        y = coord[1] + i * spacing[1]
        # Loop over each column
        for j in range(0, ncols):
            # The upper left corner is the x starting coordinate + the ROI offset * the
            # horizontal spacing between chips
            x = coord[0] + j * spacing[0]
            # calculate corners
            pt1 = [x, y]
            pt2 = [x, y + h - 1]
            pt3 = [x + w - 1, y + h - 1]
            pt4 = [x + w - 1, y]
            # Create the ROI contours and hierarchy
            rc = [np.array([[pt1], [pt2], [pt3], [pt4]], dtype=np.int32)]
            rh = np.array([[[-1, -1, -1, -1]]], dtype=np.int32)
            # Draw the rectangle on the overall mask
            overlap_img = cv2.drawContours(overlap_img, rc, -1, 255, -1)
            roi_objects.append(rc, rh)
    return roi_objects, overlap_img


def auto_grid(mask, nrows, ncols, radius=None, img=None):
    """Detect and create multiple circular ROIs on a single binary mask
    Inputs
    mask          = A binary mask.
    nrows         = Number of rows in ROI layout.
    ncols         = Number of columns in ROI layout.
    radius        = Optional parameter to specify the radius of the circular rois.
    img           = (Optional) Image from which the binary mask was created.

    Returns:
    roi_objects   = a dataclass with roi objects and hierarchies

    :param mask: numpy.ndarray
    :param nrows: int
    :param ncols: int
    :param radius: int
    :param img: numpy.ndarray
    :return roi_objects: plantcv.plantcv.classes.Objects
    """
    # Make sure the input bin_img is binary
    if len(np.unique(mask)) != 2:
        fatal_error("Input binary mask is not binary!")
    coord, spacing = _calculate_grid(mask, nrows, ncols)
    if img is None:
        img = mask
    roi_objects, overlap_img = _grid_roi(img, nrows, ncols, coord, radius, spacing)
    # Label the ROIs to check for overlap
    _, num_labels = label(overlap_img, return_num=True)
    # Check for overlapping ROIs where the number of labels is not equal to the number of expected ROIs
    if num_labels != nrows * ncols:
        warn("Two or more of the user defined regions of interest overlap! "
             "If you only see one ROI then they may overlap exactly.")
    # Draw the ROIs if requested
    # Create an array of contours and list of hierarchy for debug image
    _draw_roi(img=img, roi_contour=roi_objects)
    return roi_objects


def multi(img, coord, radius=None, spacing=None, nrows=None, ncols=None):
    """Create multiple circular ROIs on a single image

    Inputs
    img           = Input image data.
    coord         = Two-element tuple of the center of the top left object (x,y) or a list of tuples identifying
                    the center of each roi [(x1,y1),(x2,y2)]
    radius        = A single radius for all ROIs.
    spacing       = Two-element tuple of the horizontal and vertical spacing between ROIs, (x,y). Ignored if `coord`
                    is a list and `rows` and `cols` are None.
    nrows         = Number of rows in ROI layout. Should be missing or None if each center coordinate pair is listed.
    ncols         = Number of columns in ROI layout. Should be missing or None if each center coordinate pair is listed.

    Returns
    roi_objects   = a dataclass with roi objects and hierarchies

    :param img: numpy.ndarray
    :param coord: tuple, list
    :param radius: int
    :param spacing: tuple
    :param nrows: int
    :param ncols: int
    :return roi_objects: plantcv.plantcv.classes.Objects
    """
    # Grid of ROIs
    num_rois = 0
    if (isinstance(coord, tuple)) and ((nrows and ncols) is not None) and (isinstance(spacing, tuple)):
        roi_objects, overlap_img = _grid_roi(img, nrows, ncols, coord, radius, spacing)
        # The number of ROIs is the product of the number of rows and columns
        num_rois = nrows * ncols
    # User specified ROI centers
    elif (isinstance(coord, list)) and ((nrows and ncols) is None) and (spacing is None):
        roi_objects, overlap_img = _rois_from_coordinates(img=img, coord=coord, radius=radius)
        # The number of ROIs is the length of the list of coordinates
        num_rois = len(coord)
    else:
        fatal_error("Function can either make a grid of ROIs (user must provide nrows, ncols, spacing, and coord) "
                    "or take custom ROI coordinates (user must provide only a list of tuples to 'coord' parameter). "
                    "For automatic detection of a grid layout from just nrows, ncols, and a binary mask, use auto_grid")

    # Label the ROIs to check for overlap
    _, num_labels = label(overlap_img, return_num=True)
    # Check for overlapping ROIs where the number of labels is not equal to the number of expected ROIs
    if num_labels != num_rois:
        warn("Two or more of the user defined regions of interest overlap! "
             "If you only see one ROI then they may overlap exactly.")

    # Draw the ROIs if requested
    _draw_roi(img=img, roi_contour=roi_objects)
    return roi_objects


def multi_rect(img, coord, h=None, w=None, spacing=None, nrows=None, ncols=None):
    """Create multiple rectangular ROIs on a single image

    Parameters
    ----------
    img : numpy.ndarray
        Input image data.
    coord : tuple, list
        Two-element tuple of the center of the top left object (x,y) or a list of tuples identifying
        the top left corner of each roi [(x1,y1),(x2,y2), ...]
    h : int, optional
        The height of each rectangular ROI, by default None
    w : int, optional
        The width of each rectangular ROI, by default None
    spacing : tuple, optional
        Two-element tuple of the horizontal and vertical spacing between ROIs, (x,y). Ignored if `coord`
        is a list and `rows` and `cols` are None.
    nrows : int, optional
        Number of rows in ROI layout. Should be missing or None if each center coordinate pair is listed.
    ncols : int, optional
        Number of columns in ROI layout. Should be missing or None if each center coordinate pair is listed.

    Returns
    -------
    plantcv.plantcv.classes.Objects
        A dataclass with roi objects and hierarchies
    """
    # Grid of ROIs
    num_rois = 0
    if (isinstance(coord, tuple)) and ((nrows and ncols) is not None) and (isinstance(spacing, tuple)):
        roi_objects, overlap_img = _grid_roi_rect(img, nrows, ncols, h, w, coord, spacing)
        # The number of ROIs is the product of the number of rows and columns
        num_rois = nrows * ncols
    # User specified ROI centers
    elif (isinstance(coord, list)) and ((nrows and ncols) is None) and (spacing is None):
        roi_objects, overlap_img = _rect_rois_from_coordinates(img=img, h=h, w=w, coord=coord)
        # The number of ROIs is the length of the list of coordinates
        num_rois = len(coord)
    else:
        fatal_error("Function can either make a grid of ROIs (user must provide nrows, ncols, spacing, and coord) "
                    "or take custom ROI coordinates (user must provide only a list of tuples to 'coord' parameter). "
                    "For automatic detection of a grid layout from just nrows, ncols, and a binary mask, use auto_grid")

    # Label the ROIs to check for overlap
    _, num_labels = label(overlap_img, return_num=True)
    # Check for overlapping ROIs where the number of labels is not equal to the number of expected ROIs
    if num_labels != num_rois:
        warn("Two or more of the user defined regions of interest overlap! "
             "If you only see one ROI then they may overlap exactly.")

    # Draw the ROIs if requested
    _draw_roi(img=img, roi_contour=roi_objects)
    return roi_objects


def auto_wells(gray_img, mindist, candec, accthresh, minradius, maxradius, nrows, ncols, radiusadjust=None, roi=None):
    """Hough Circle Well Detection.

    Keyword inputs:
    gray_img = gray image (np.ndarray)
    mindist = minimum distance between detected circles
    candec = higher threshold of canny edge detector
    accthresh = accumulator threshold for the circle centers
    minradius = minimum circle radius
    maxradius = maximum circle radius
    nrows = expected number of rows
    ncols = expected number of columns
    radiusadjust = amount to adjust the average radius, this can be desirable
    if you want ROI to sit inside a well, for example (in that case you might
    set it to a negative value).
    roi = optional rectangular ROI to find wells within.

    :param gray_img: np.ndarray
    :param mindist: int
    :param candec: int
    :param accthresh: int
    :param minradius: int
    :param maxradius: int
    :param nrows = int
    :param ncols = int
    :param roi = plantcv.plantcv.classes.Objects
    :return roi: plantcv.plantcv.classes.Objects
    """
    # Use hough circle helper function
    maxfind = nrows * ncols
    df, sub_img = _rect_filter(gray_img, roi, _hough_circle,
                               **{"mindist": mindist, "candec": candec,
                                  "accthresh": accthresh, "minradius": minradius,
                                  "maxradius": maxradius, "maxfound": maxfind})
    # cast gray image to color
    cimg = cv2.cvtColor(gray_img, cv2.COLOR_GRAY2BGR)
    # slice ROI back into original image
    img = _rect_replace(cimg, sub_img, roi)
    # adjust data.frame x/y points to match full scale image if ROI was used
    if roi is not None:
        xstart = roi.contours[0][0][0][0][0].astype("int32")
        ystart = roi.contours[0][0][0][0][1].astype("int32")
        df['x'] = df['x'] + xstart
        df['y'] = df['y'] + ystart

    _debug(img, filename=os.path.join(params.debug_outdir, str(params.device) + '_roi_houghcircle.png'), cmap='gray')

    xlist = []
    centers_x = df['x'].values.reshape(-1, 1)
    centers_y = df['y'].values.reshape(-1, 1)
    gm_x = GaussianMixture(n_components=ncols, random_state=0).fit(centers_x)
    gm_y = GaussianMixture(n_components=nrows, random_state=0).fit(centers_y)
    clusters_x = gm_x.means_[:, 0]
    clusters_y = gm_y.means_[:, 0]

    sorted_indicesx = np.argsort(clusters_x)
    sorted_x = np.sort(clusters_x)
    xlist = list(range(len(clusters_x)))
    clusterxdf = {'clusters_x': sorted_x, 'sorted': sorted_indicesx,
                  'xindex': [0]*len(clusters_x)}
    xdf = pd.DataFrame(clusterxdf)
    xdf = xdf.sort_values(by='clusters_x', ascending=True)
    xdf['xindex'] = xlist

    sorted_indicesy = np.argsort(clusters_y)
    sorted_y = np.sort(clusters_y)
    ylist = list(range(len(clusters_y)))
    clusterydf = {'clusters_y': sorted_y, 'sorted': sorted_indicesy,
                  'yindex': [0]*len(clusters_y)}
    ydf = pd.DataFrame(clusterydf)
    ydf = ydf.sort_values(by='clusters_y', ascending=True)
    ydf['yindex'] = ylist

    df['column'] = gm_x.predict(centers_x)
    df['row'] = gm_y.predict(centers_y)

    df['row'] = df['row'].replace(list(ydf['sorted']), list(ydf['yindex']))
    df['column'] = df['column'].replace(list(xdf['sorted']), list(xdf['xindex']))

    df = df.sort_values(by=['row', 'column'], ascending=[True, True])
    df['xy'] = list(zip(df.x, df.y))

    radiusfinal = int(np.mean(list(df['radius'])))+radiusadjust

    rois = multi(gray_img, list(df['xy']), radius=radiusfinal)

    return rois


def custom(img, vertices):
    """Create an custom polygon ROI.

    Inputs:
    img = An RGB or grayscale image to plot the ROI on in debug mode.
    vertices = List of vertices of the desired polygon ROI

    Outputs:
    roi = a dataclass with the roi object and hierarchy

    :param img: numpy.ndarray
    :param vertices: list
    :return roi: plantcv.plantcv.classes.Objects
    """
    # Get the height and width of the reference image
    height, width = np.shape(img)[:2]

    # Check that the ROI doesn't go off the screen
    for i in vertices:
        (x, y) = i
        if x < 0 or x > width or y < 0 or y > height:
            fatal_error("An ROI extends outside of the image!")

    roi_contour = [np.array(vertices, dtype=np.int32)]
    roi_hierarchy = np.array([[[-1, -1, -1, -1]]], dtype=np.int32)
    roi = Objects(contours=[roi_contour], hierarchy=[roi_hierarchy])

    # Draw the ROIs if requested
    _draw_roi(img=img, roi_contour=roi)

    return roi


# Filter a mask based on a region of interest
def filter(mask, roi, roi_type="partial"):
    """
    Filter a mask using a region of interest.

    Connected regions of non-zero pixels outside the ROI are set to zero.

    Parameters
    ----------
    mask : numpy.ndarray
        Binary image data to be filtered.
    roi : plantcv.plantcv.classes.Objects
        Region of interest, an instance of the Object class output from a ROI function.
    roi_type : str, optional
        Type of ROI filtering: 'cutto', 'partial' (default, for partially inside),
        'largest' (keep only the largest contour), or 'within'.

    Returns
    -------
    filtered_mask : numpy.ndarray
        Mask image after ROI filtering.
    """
    found_obj, found_hier = _cv2_findcontours(bin_img=mask)

    _, _, filtered_mask = _roi_filter(img=mask, roi=roi, obj=found_obj,
                                      hierarchy=found_hier, roi_type=roi_type)

    _debug(filtered_mask, filename=os.path.join(params.debug_outdir, str(params.device) + '_roi_filter.png'), cmap='gray')

    return filtered_mask



"""Analyzes the spectral reflectance values of objects in an image."""
import os
import numpy as np
from plantcv.plantcv import outputs, params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _iterate_analysis


def spectral_reflectance(hsi, labeled_mask, n_labels=1, label=None):
    """Analyze spectral reflectance values of objects in an image.

    Inputs:
    hsi          = Hyperspectral image data
    labeled_mask = Labeled mask of objects (32-bit).
    n_labels     = Total number expected individual objects (default = 1).
    label        = Optional label parameter, modifies the variable name of observations recorded (default = "default").

    Returns:
    spectral_chart = Spectral reflectance plot.

    :param hsi: plantcv.plantcv.Spectral_data
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param label: str
    :return spectral_chart: altair.vegalite.v5.api.FacetChart
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    _ = _iterate_analysis(img=hsi, labeled_mask=labeled_mask, n_labels=n_labels, label=label, function=_analyze_spectral)
    spectral_chart = outputs.plot_dists(variable="wavelength_means")
    _debug(visual=spectral_chart, filename=os.path.join(params.debug_outdir, str(params.device) + '_mean_reflectance.png'))
    return spectral_chart


def _analyze_spectral(img, mask, label):
    """This extracts the hyperspectral reflectance values of each pixel.

    Inputs:
    img          = Hyperspectral data instance
    mask         = Binary mask made from selected contours
    label        = optional label parameter, modifies the variable name of observations recorded

    Returns:
    analysis_img = output image

    :param array: __main__.Spectral_data
    :param mask: numpy array
    :param histplot: bool
    :param label: str
    :return img: plantcv.plantcv.Spectral_data
    """
    array_data = img.array_data

    # List of wavelengths recorded created from parsing the header file will be string, make list of floats
    wavelength_data = array_data[np.where(mask > 0)]

    # Initialize analysis output values with zeros
    wavelength_means = np.full(len(img.wavelength_dict), 0)
    max_per_band = np.full(len(img.wavelength_dict), 0)
    min_per_band = np.full(len(img.wavelength_dict), 0)
    std_per_band = np.full(len(img.wavelength_dict), 0)

    avg_reflectance = 0
    std_reflectance = 0
    median_reflectance = 0

    # check if mask is empty and procees only if it is not
    if wavelength_data.size != 0:
        # Calculate mean reflectance across wavelengths
        wavelength_means = wavelength_data.mean(axis=0)
        max_per_band = wavelength_data.max(axis=0)
        min_per_band = wavelength_data.min(axis=0)
        std_per_band = wavelength_data.std(axis=0)

        # Calculate reflectance statistics
        avg_reflectance = np.average(wavelength_data)
        std_reflectance = np.std(wavelength_data)
        median_reflectance = np.median(wavelength_data)

    # Create lists with wavelengths in float format rather than as strings
    # and make a list of the frequencies since they are in an array
    new_wavelengths = []
    band_averages = []
    new_std_per_band = []
    new_max_per_band = []
    new_min_per_band = []

    for i, wavelength in enumerate(img.wavelength_dict):
        new_wavelengths.append(wavelength)
        band_averages.append((wavelength_means[i]).astype(float))
        new_std_per_band.append(std_per_band[i].astype(float))
        new_max_per_band.append(max_per_band[i].astype(float))
        new_min_per_band.append(min_per_band[i].astype(float))

    wavelength_labels = []
    for i in img.wavelength_dict.keys():
        wavelength_labels.append(i)

    # Store data into outputs class
    outputs.add_observation(sample=label, variable='global_mean_reflectance', trait='global mean reflectance',
                            method='plantcv.plantcv.analyze.spectral_reflectance', scale='reflectance',
                            datatype=float, value=float(avg_reflectance), label='reflectance')
    outputs.add_observation(sample=label, variable='global_median_reflectance', trait='global median reflectance',
                            method='plantcv.plantcv.analyze.spectral_reflectance', scale='reflectance',
                            datatype=float, value=float(median_reflectance), label='reflectance')
    outputs.add_observation(sample=label, variable='global_spectral_std',
                            trait='pixel-wise standard deviation per band',
                            method='plantcv.plantcv.analyze.spectral_reflectance', scale='None', datatype=float,
                            value=float(std_reflectance), label='reflectance')
    outputs.add_observation(sample=label, variable='wavelength_means', trait='mean reflectance per band',
                            method='plantcv.plantcv.analyze.spectral_reflectance', scale='reflectance', datatype=list,
                            value=band_averages, label=wavelength_labels)
    outputs.add_observation(sample=label, variable='max_reflectance', trait='maximum reflectance per band',
                            method='plantcv.plantcv.analyze.spectral_reflectance', scale='reflectance', datatype=list,
                            value=new_max_per_band, label=wavelength_labels)
    outputs.add_observation(sample=label, variable='min_reflectance', trait='minimum reflectance per band',
                            method='plantcv.plantcv.analyze.spectral_reflectance', scale='reflectance', datatype=list,
                            value=new_min_per_band, label=wavelength_labels)
    outputs.add_observation(sample=label, variable='spectral_std', trait='pixel-wise standard deviation per band',
                            method='plantcv.plantcv.analyze.spectral_reflectance', scale='None', datatype=list,
                            value=new_std_per_band, label=wavelength_labels)
    return img



from plantcv.plantcv.photosynthesis.read_cropreporter import read_cropreporter
from plantcv.plantcv.photosynthesis.reassign_frame_labels import reassign_frame_labels

__all__ = ["read_cropreporter", "reassign_frame_labels"]



# Warnings handling

import sys
from plantcv.plantcv import params


def warn(warning):
    """Print a warning message to stderr.

    Parameters
    ----------
    warning : str
        Warning message to print.

    Returns
    -------
    None
        Function does not return anything.
    """
    if params.verbose:
        print(f"Warning: {warning}", file=sys.stderr)



# Break skeleton into segments

import os
import cv2
from plantcv.plantcv import params
from plantcv.plantcv import color_palette
from plantcv.plantcv.morphology import find_branch_pts
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _dilate, _image_subtract


def segment_skeleton(skel_img, mask=None):
    """Segment a skeleton image into pieces.

    Inputs:
    skel_img         = Skeletonized image
    mask             = (Optional) binary mask for debugging. If provided, debug image will be overlaid on the mask.

    Returns:
    segmented_img       = Segmented debugging image
    segment_objects     = list of contours

    :param skel_img: numpy.ndarray
    :param mask: numpy.ndarray
    :return segmented_img: numpy.ndarray
    :return segment_objects: list
    """
    # Find branch points
    bp = find_branch_pts(skel_img)
    bp = _dilate(bp, 3, 1)

    # Subtract from the skeleton so that leaves are no longer connected
    segments = _image_subtract(skel_img, bp)

    # Gather contours of leaves
    segment_objects, _ = _cv2_findcontours(bin_img=segments)

    # Color each segment a different color, do not used a previously saved color scale
    rand_color = color_palette(num=len(segment_objects), saved=False)

    if mask is None:
        segmented_img = skel_img.copy()
    else:
        segmented_img = mask.copy()

    segmented_img = cv2.cvtColor(segmented_img, cv2.COLOR_GRAY2RGB)
    for i, _ in enumerate(segment_objects):
        cv2.drawContours(segmented_img, segment_objects, i, rand_color[i], params.line_thickness, lineType=8)

    _debug(visual=segmented_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segmented.png"))

    return segmented_img, list(segment_objects)



# RGB -> LAB -> Gray

import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params
from plantcv.plantcv._helpers import _rgb2lab


def rgb2gray_lab(rgb_img, channel):
    """Convert image from RGB colorspace to LAB colorspace. Returns the specified subchannel as a gray image.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data
    channel : str
        color subchannel (l = lightness, a = green-magenta, b = blue-yellow)

    Returns
    -------
    numpy.ndarray
        grayscale image from one LAB color channel
    """
    # Convert RGB to LAB and return the specified subchannel as a gray image
    gray_img = _rgb2lab(rgb_img=rgb_img, channel=channel)

    # The allowable channel inputs are l, a or b
    names = {"l": "lightness", "a": "green-magenta", "b": "blue-yellow"}

    # Display debug image
    _debug(visual=gray_img,
           filename=os.path.join(params.debug_outdir, f"{params.device}_lab_{names[channel.lower()]}.png"), cmap="gray")

    return gray_img



"""Analyze the horizontal distribution of the plant relative to a vertical reference line."""
import os
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _iterate_analysis, _grayscale_to_rgb, _scale_size
from plantcv.plantcv.analyze.bound_horizontal import _get_boundary_values, _boundary_img_annotation
from plantcv.plantcv import params
from plantcv.plantcv import outputs


def bound_vertical(img, labeled_mask, line_position, n_labels=1, label=None):
    """
    Analyze the horizontal distribution of the plant relative to a vertical reference line for individual objects.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting.
    labeled_mask : numpy.ndarray
        Labeled mask of objects (32-bit).
    line_position : int
        Position of boundary line in pixels from left to right (a value of 0 draws the line through the left of the image).
    n_labels : int, optional
        Total number of expected individual objects (default = 1).
    label : str, optional
        Optional label parameter, modifies the variable name of observations recorded (default = params.sample_label).

    Returns
    -------
    analysis_image : numpy.ndarray
        Diagnostic image showing measurements.
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    img = _iterate_analysis(img=img, labeled_mask=labeled_mask, n_labels=n_labels,
                            label=label, function=_analyze_bound_vertical,
                            **{"line_position": line_position})
    img = _boundary_img_annotation(img, labeled_mask, line_position, 1)
    # Debugging
    _debug(visual=img, filename=os.path.join(params.debug_outdir, str(params.device) + '_boundary_on_img.png'))
    return img


def _analyze_bound_vertical(img, mask, line_position, label):
    """
    Analyze the mask relative to a user-input vertical boundary line.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting.
    mask : numpy.ndarray
        Binary mask made from selected contours.
    line_position : int
        Position of boundary line in pixels from left to right (a value of 0 draws the line through the left of the image).
    label : str
        Optional label parameter, modifies the variable name of observations recorded.

    Returns
    -------
    ori_img : numpy.ndarray
        Output image.
    """
    # Initialize output measurements
    width_left_bound = 0
    width_right_bound = 0
    left_bound_area = 0
    percent_bound_area_left = 0
    right_bound_area = 0
    percent_bound_area_right = 0

    ori_img = np.copy(img)
    # Skip empty masks
    if np.count_nonzero(mask) != 0 and line_position >= 0:
        # Draw line horizontal line through bottom of image, that is adjusted to user input height
        ori_img = _grayscale_to_rgb(img)
        left_mask = np.copy(mask).astype(bool)
        right_mask = np.copy(mask).astype(bool)
        left_mask[:, line_position:np.shape(left_mask)[1] + 1] = np.zeros(
            np.shape(left_mask[:, line_position:np.shape(left_mask)[1] + 1]))
        right_mask[:, 0:line_position - 1] = np.zeros(
            np.shape(right_mask[:, 0:line_position - 1]))
        tot_area = np.sum(mask.astype(bool))
        left_bound_area, width_left_bound, percent_bound_area_left = _get_boundary_values(left_mask, tot_area, 1)
        right_bound_area, width_right_bound, percent_bound_area_right = _get_boundary_values(right_mask, tot_area, 1)

    outputs.add_observation(sample=label, variable='vertical_reference_position', trait='vertical reference position',
                            method='plantcv.plantcv.analyze.bound_vertical', scale='none', datatype=int,
                            value=line_position, label='none')
    outputs.add_observation(sample=label, variable='width_left_reference', trait='width left of reference',
                            method='plantcv.plantcv.analyze.bound_vertical', scale=params.unit, datatype=int,
                            value=_scale_size(width_left_bound), label=params.unit)
    outputs.add_observation(sample=label, variable='width_right_reference', trait='width right of reference',
                            method='plantcv.plantcv.analyze.bound_vertical', scale=params.unit, datatype=int,
                            value=_scale_size(width_right_bound), label=params.unit)
    outputs.add_observation(sample=label, variable='area_left_reference', trait='area left of reference',
                            method='plantcv.plantcv.analyze.bound_vertical', scale=params.unit, datatype=int,
                            value=_scale_size(left_bound_area, "area"), label=params.unit)
    outputs.add_observation(sample=label, variable='percent_area_left_reference',
                            trait='percent area left of reference', method='plantcv.plantcv.analyze.bound_vertical',
                            scale='none', datatype=float,
                            value=percent_bound_area_left, label='none')
    outputs.add_observation(sample=label, variable='area_right_reference', trait='area right of reference',
                            method='plantcv.plantcv.analyze.bound_vertical', scale=params.unit, datatype=int,
                            value=_scale_size(right_bound_area, "area"), label=params.unit)
    outputs.add_observation(sample=label, variable='percent_area_right_reference',
                            trait='percent area right of reference', method='plantcv.plantcv.analyze.bound_vertical',
                            scale='none', datatype=float, value=percent_bound_area_right, label='none')

    return ori_img



from plantcv.plantcv.homology.acute import acute
from plantcv.plantcv.homology.space import space
from plantcv.plantcv.homology.starscape import starscape
from plantcv.plantcv.homology.constella import constella
from plantcv.plantcv.homology.constellaqc import constellaqc
from plantcv.plantcv.homology.x_axis_pseudolandmark import x_axis_pseudolandmarks
from plantcv.plantcv.homology.y_axis_pseudolandmarks import y_axis_pseudolandmarks
from plantcv.plantcv.homology.landmark_reference_pt_dist import landmark_reference_pt_dist
from plantcv.plantcv.homology.scale_features import scale_features

__all__ = ["acute", "space", "starscape", "constella", "constellaqc", "x_axis_pseudolandmarks", "y_axis_pseudolandmarks",
           "landmark_reference_pt_dist", "scale_features"]



import cv2
import numpy as np
import math
from skimage import morphology
from plantcv.plantcv import fatal_error, warn
from plantcv.plantcv import params
import pandas as pd


def _closing(gray_img, kernel=None, roi=None):
    """Wrapper for scikit-image closing functions.

    Opening can remove small dark spots (i.e. pepper).

    Parameters
    ----------
    gray_img = numpy.ndarray
             input image (grayscale or binary)
    kernel   = numpy.ndarray
             optional neighborhood, expressed as an array of 1s and 0s. If None, use cross-shaped structuring element.
    roi : Objects class
             Optional rectangular ROI to erode within

    Returns
    -------
    numpy.ndarray
         filtered (holes closed) image

    Raises
    ------
    ValueError
        If input image is not gray-scale
    """
    # Make sure the image is binary/grayscale
    if len(np.shape(gray_img)) != 2:
        fatal_error("Input image must be grayscale or binary")

    if len(np.unique(gray_img)) <= 2:
        bool_img = gray_img.astype(bool)
        sub_img = _rect_filter(bool_img, roi=roi, function=morphology.binary_closing,
                               **{"footprint": kernel})
        filtered_img = sub_img.astype(np.uint8) * 255
        replaced_img = _rect_replace(bool_img.astype(np.uint8) * 255, filtered_img, roi)
    # Otherwise use method appropriate for grayscale images
    else:
        filtered_img = _rect_filter(gray_img,
                                    roi=roi,
                                    function=morphology.closing,
                                    **{"footprint": kernel})
        replaced_img = _rect_replace(gray_img, filtered_img, roi)

    return replaced_img


def _image_subtract(gray_img1, gray_img2):
    """Subtract values of one gray-scale image array from another gray-scale image array.

    The
    resulting gray-scale image array has a minimum element value of zero. That is all negative values resulting from the
    subtraction are forced to zero.

    Parameters
    ----------
    gray_img1 : numpy.ndarray
              Grayscale image data from which gray_img2 will be subtracted
    gray_img2 : numpy.ndarray
              Grayscale image data which will be subtracted from gray_img1

    Returns
    -------
    new_img = subtracted image

    Raises
    ------
    ValueError
         If input image is not gray scale
    """
    # check inputs for gray-scale
    if len(np.shape(gray_img1)) != 2 or len(np.shape(gray_img2)) != 2:
        fatal_error("Input image is not gray-scale")

    new_img = gray_img1.astype(np.float64) - gray_img2.astype(np.float64)  # subtract values
    new_img[np.where(new_img < 0)] = 0  # force negative array values to zero
    new_img = new_img.astype(np.uint8)  # typecast image to 8-bit image

    return new_img  # return


def _erode(gray_img, ksize, i, roi=None):
    """Perform morphological 'erosion' filtering.

    Keeps pixel in center of the kernel if conditions set in kernel are
       true, otherwise removes pixel.

    Parameters
    ----------
    gray_img : numpy.ndarray
             Grayscale (usually binary) image data
    ksize : int
             Kernel size (int). A ksize x ksize kernel will be built. Must be greater than 1 to have an effect.
    i : int
             interations, i.e. number of consecutive filtering passes
    roi : Objects class
             Optional rectangular ROI to erode within

    Returns
    -------
    numpy.ndarray
         Eroded result image

    Raises
    ------
    ValueError
        If ksize is less than or equal to 1.
    """
    if ksize <= 1:
        raise ValueError('ksize needs to be greater than 1 for the function to have an effect')

    kernel1 = int(ksize)
    kernel2 = np.ones((kernel1, kernel1), np.uint8)
    sub_er_img = _rect_filter(img=gray_img, roi=roi, function=cv2.erode,
                              **{"kernel": kernel2, "iterations": i})
    er_img = _rect_replace(gray_img, sub_er_img, roi)

    return er_img


def _dilate(gray_img, ksize, i, roi=None):
    """Performs morphological 'dilation' filtering.

    Parameters
    ----------
    gray_img : numpy.ndarray
        Grayscale image data to be dilated
    ksize : int
        Kernel size (int). A k x k kernel will be built. Must be greater than 1 to have an effect.
    i : int
        Number of iterations (i.e. how many times to apply the dilation).
    roi : Objects class
        Optional rectangular ROI to erode within

    Returns
    -------
    numpy.ndarray
        Dilation result image

    Raises
    ------
    ValueError
        If ksize is less than or equal to 1.
    """
    if ksize <= 1:
        raise ValueError('ksize needs to be greater than 1 for the function to have an effect')

    kernel1 = int(ksize)
    kernel2 = np.ones((kernel1, kernel1), np.uint8)
    sub_dil_img = _rect_filter(img=gray_img, roi=roi, function=cv2.dilate,
                               **{"kernel": kernel2, "iterations": i})
    dil_img = _rect_replace(gray_img, sub_dil_img, roi)

    return dil_img


def _find_segment_ends(skel_img, leaf_objects, plotting_img, size):
    """Find both segment ends and sort into tips or inner branchpoints.
    Inputs:
    skel_img         = Skeletonized image
    leaf_objects     = List of leaf segments
    plotting_img     = Mask for debugging, might be a copy of the Skeletonized image
    size             = Size of inner segment ends (in pixels)
    :param skel_img: numpy.ndarray
    :param leaf_objects: list
    :param plotting_img: numpy.ndarray
    """
    labeled_img = cv2.cvtColor(plotting_img, cv2.COLOR_GRAY2RGB)
    tips, _, _ = _find_tips(skel_img)
    # Initialize list of tip data points
    labels = []
    tip_list = []
    inner_list = []
    remove = []
    sortabled_objs = leaf_objects.copy()

    # Find segment end coordinates
    for i in range(len(leaf_objects)):
        labels.append(i)
        # Draw leaf objects
        find_segment_tangents = np.zeros(labeled_img.shape[:2], np.uint8)
        cv2.drawContours(find_segment_tangents, leaf_objects, i, 255, 1, lineType=8)
        cv2.drawContours(labeled_img, leaf_objects, i, (150, 150, 150), params.line_thickness, lineType=8)  # segments debug
        # Prune back ends of leaves
        pruned_segment = _iterative_prune(find_segment_tangents, size)
        # Segment ends are the portions pruned off
        ends = find_segment_tangents - pruned_segment
        segment_end_obj, _ = _cv2_findcontours(bin_img=ends)
        branch_pt_found = False
        coords = []
        # Determine if a segment is segment tip or branch point
        for j, obj in enumerate(segment_end_obj):
            segment_plot = np.zeros(skel_img.shape[:2], np.uint8)
            cv2.drawContours(segment_plot, obj, -1, 255, 1, lineType=8)
            segment_plot = _dilate(segment_plot, 3, 1)
            overlap_img = _logical_operation(segment_plot, tips, 'and')
            x, y = segment_end_obj[j].ravel()[:2]
            coord = (int(x), int(y))
            coords.append(coord)
            # If none of the tips are within a segment_end then it's an insertion segment
            if np.sum(overlap_img) == 0:
                inner_list.append(coord)
                cv2.circle(labeled_img, coord, params.line_thickness, (50, 0, 255), -1)  # Red auricles/branch points
                branch_pt_found = True
            else:
                tip_list.append(coord)
                cv2.circle(labeled_img, coord, params.line_thickness, (0, 255, 0), -1)  # green tips
        if not branch_pt_found:  # there is no branch point associated with a given segment and therefore it cannot be sorted
            remove.append(i)
            # Plot the ends if found
            if len(coords) > 1:
                # Plot the tip that is closest to the stem
                x_min, y_min, w, h = cv2.boundingRect(skel_img)
                cx = int((x_min + (w / 2)))
                cy = int(y_min + h)
                dist0 = math.dist(coords[0], (cx, cy))
                dist1 = math.dist(coords[1], (cx, cy))
                m = 1
                if dist0 < dist1:
                    m = 0

                cv2.circle(labeled_img, (cx, cy), params.line_thickness + 10, (255, 0, 0), 5)  # estimated centroid point
                cv2.circle(labeled_img, coords[m], params.line_thickness, (255, 20, 20), -1)  # estimated sorting point
                cv2.putText(img=labeled_img, text=str(int(dist0)), org=coords[0], fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                            fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)
                cv2.putText(img=labeled_img, text=str(int(dist1)), org=coords[1], fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                            fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)

    # Remove the segments that cannot be resorted, since they do not have a branch point
    for k in remove:
        sortabled_objs.pop(k)

    return labeled_img, tip_list, inner_list, labels, sortabled_objs


def _iterative_prune(skel_img, size):
    """Iteratively remove endpoints (tips) from a skeletonized image.
    The pruning algorithm was inspired by Jean-Patrick Pommier: https://gist.github.com/jeanpat/5712699
    "Prunes" barbs off a skeleton.
    Inputs:
    skel_img    = Skeletonized image
    size        = Size to get pruned off each branch
    Returns:
    pruned_img  = Pruned image
    :param skel_img: numpy.ndarray
    :param size: int
    :return pruned_img: numpy.ndarray
    """
    pruned_img = skel_img.copy()
    # Store debug
    debug = params.debug
    params.debug = None

    # Iteratively remove endpoints (tips) from a skeleton
    for _ in range(0, size):
        endpoints, _, _ = _find_tips(pruned_img)
        pruned_img = _image_subtract(pruned_img, endpoints)

    # Make debugging image
    pruned_plot = np.zeros(skel_img.shape[:2], np.uint8)
    pruned_plot = cv2.cvtColor(pruned_plot, cv2.COLOR_GRAY2RGB)
    skel_obj, skel_hierarchy = _cv2_findcontours(bin_img=pruned_img)
    pruned_obj, pruned_hierarchy = _cv2_findcontours(bin_img=pruned_img)

    # Reset debug mode
    params.debug = debug

    cv2.drawContours(pruned_plot, skel_obj, -1, (0, 0, 255), params.line_thickness,
                     lineType=8, hierarchy=skel_hierarchy)
    cv2.drawContours(pruned_plot, pruned_obj, -1, (255, 255, 255), params.line_thickness,
                     lineType=8, hierarchy=pruned_hierarchy)

    return pruned_img


def _find_tips(skel_img, mask=None):
    """Helper function to find tips in skeletonized image.
    The endpoints algorithm was inspired by Jean-Patrick Pommier: https://gist.github.com/jeanpat/5712699

    Inputs:
    skel_img    = Skeletonized image
    mask        = (Optional) binary mask for debugging. If provided, debug image will be overlaid on the mask.
    label       = Optional label parameter, modifies the variable name of
                  observations recorded (default = pcv.params.sample_label).
    Returns:
    tip_img   = Image with just tips, rest 0

    :param skel_img: numpy.ndarray
    :param mask: numpy.ndarray
    :param label: str
    :return tip_img: numpy.ndarray
    """
    # In a kernel: 1 values line up with 255s, -1s line up with 0s, and 0s correspond to dont care
    endpoint1 = np.array([[-1, -1, -1],
                          [-1, 1, -1],
                          [0, 1,  0]])
    endpoint2 = np.array([[-1, -1, -1],
                          [-1, 1, 0],
                          [-1, 0, 1]])

    endpoint3 = np.rot90(endpoint1)
    endpoint4 = np.rot90(endpoint2)
    endpoint5 = np.rot90(endpoint3)
    endpoint6 = np.rot90(endpoint4)
    endpoint7 = np.rot90(endpoint5)
    endpoint8 = np.rot90(endpoint6)

    endpoints = [endpoint1, endpoint2, endpoint3, endpoint4, endpoint5, endpoint6, endpoint7, endpoint8]
    tip_img = np.zeros(skel_img.shape[:2], dtype=int)
    for endpoint in endpoints:
        tip_img = np.logical_or(cv2.morphologyEx(skel_img, op=cv2.MORPH_HITMISS, kernel=endpoint,
                                                 borderType=cv2.BORDER_CONSTANT, borderValue=0), tip_img)
    tip_img = tip_img.astype(np.uint8) * 255
    # Store debug
    debug = params.debug
    params.debug = None
    tip_objects, _ = _cv2_findcontours(bin_img=tip_img)

    if mask is None:
        # Make debugging image
        dilated_skel = _dilate(skel_img, params.line_thickness, 1)
        tip_plot = cv2.cvtColor(dilated_skel, cv2.COLOR_GRAY2RGB)

    else:
        # Make debugging image on mask
        mask_copy = mask.copy()
        tip_plot = cv2.cvtColor(mask_copy, cv2.COLOR_GRAY2RGB)
        skel_obj, skel_hier = _cv2_findcontours(bin_img=skel_img)
        cv2.drawContours(tip_plot, skel_obj, -1, (150, 150, 150), params.line_thickness,
                         lineType=8, hierarchy=skel_hier)

    # Initialize list of tip data points
    tip_list = []
    tip_labels = []
    for i, tip in enumerate(tip_objects):
        x, y = tip.ravel()[:2]
        coord = (int(x), int(y))
        tip_list.append(coord)
        tip_labels.append(i)
        cv2.circle(tip_plot, (x, y), params.line_thickness, (0, 255, 0), -1)

    # Reset debug mode
    params.debug = debug

    return tip_img, tip_list, tip_labels


def _hough_circle(gray_img, mindist, candec, accthresh, minradius, maxradius, maxfound=None):
    """
    Hough Circle Detection

    Keyword inputs:
    gray_img = gray image (np.ndarray)
    mindist = minimum distance between detected circles
    candec = higher threshold of canny edge detector
    accthresh = accumulator threshold for the circl centers
    minradius = minimum circle radius
    maxradius = maximum circle radius
    maxfound = maximum number of circles to find

    :param gray_img: np.ndarray
    :param mindist: int
    :param candec: int
    :param accthresh: int
    :param minradius: int
    :param maxradius: int
    :param maxfound: None or int
    :return dataframe: pandas dataframe
    :return img: np.ndarray
    """
    # Store debug
    debug = params.debug
    params.debug = None

    circles = cv2.HoughCircles(gray_img, cv2.HOUGH_GRADIENT,
                               dp=1, minDist=mindist,
                               param1=candec, param2=accthresh,
                               minRadius=minradius, maxRadius=maxradius)

    cimg = cv2.cvtColor(gray_img, cv2.COLOR_GRAY2BGR)
    x = []
    y = []
    radius = []
    if circles is None:
        fatal_error('number of circles found is None with these parameters')
    circles = np.uint16(np.around(circles))
    if maxfound is not None:
        if maxfound >= len(circles[0, :]):
            for i in circles[0, :]:
                # draw the outer circle
                cv2.circle(cimg, (i[0], i[1]), i[2], (0, 255, 0),
                           params.line_thickness)
                # draw the center of the circle
                cv2.circle(cimg, (i[0], i[1]), 2, (0, 0, 255),
                           params.line_thickness)
                x.append(i[0])
                y.append(i[1])
                radius.append(i[2])
        else:
            for n, i in enumerate(circles[0, :]):
                if n <= (maxfound-1):
                    # draw the outer circle
                    cv2.circle(cimg, (i[0], i[1]), i[2], (0, 255, 0),
                               params.line_thickness)
                    # draw the center of the circle
                    cv2.circle(cimg, (i[0], i[1]), 2, (0, 0, 255),
                               params.line_thickness)
                    x.append(i[0])
                    y.append(i[1])
                    radius.append(i[2])
            warn('Number of found circles is ' +
                 str(len(circles[0, :])) +
                 ' Change Parameters. Only drawing first '+str(maxfound))
    else:
        for i in circles[0, :]:
            # draw the outer circle
            cv2.circle(cimg, (i[0], i[1]), i[2], (0, 255, 0),
                       params.line_thickness)
            # draw the center of the circle
            cv2.circle(cimg, (i[0], i[1]), 2, (0, 0, 255),
                       params.line_thickness)
            x.append(i[0])
            y.append(i[1])
            radius.append(i[2])

    data = {'x': x, 'y': y, 'radius': radius}
    df = pd.DataFrame(data)

    # Reset debug mode
    params.debug = debug

    return df, cimg


def _cv2_findcontours(bin_img):
    """
    Helper function for OpenCV findContours.

    Reduces duplication of calls to findContours in the event the OpenCV function changes.

    Keyword inputs:
    bin_img = Binary image (np.ndarray)

    :param bin_img: np.ndarray
    :return contours: list
    :return hierarchy: np.array
    """
    contours, hierarchy = cv2.findContours(np.copy(bin_img), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)[-2:]

    return contours, hierarchy


def _roi_filter(img, roi, obj, hierarchy, roi_type="partial"):
    """
    Helper function to filter contours using a single ROI.

    Finds objects partially inside a region of interest or cuts objects to the ROI.

    Parameters
    ----------
    img : numpy.ndarray
        RGB, binary, or grayscale image data for shape.
    roi : plantcv.plantcv.classes.Objects
        Region of interest, an instance of the Object class output from a ROI function.
    obj : list
        Contours of objects, output from "_cv2_findcontours" function.
    hierarchy : numpy.ndarray
        Hierarchy of objects, output from "_cv2_findcontours" function.
    roi_type : str, optional
        Type of ROI filtering. Options are:
        - 'partial': Find objects partially inside the ROI (default).
        - 'cutto': Cut objects to the ROI.
        - 'largest': Keep only the largest contour.
        - 'within': Keep only objects fully within the ROI.

    Returns
    -------
    kept_cnt : list
        List of kept contours after filtering.
    kept_hier : numpy.ndarray
        Hierarchy of kept contours.
    mask : numpy.ndarray
        Mask image showing the filtered contours.

    Raises
    ------
    RuntimeError
        If an invalid `roi_type` is provided.

    Notes
    -----
    If a multi-ROI is provided, only the first ROI will be used. For multi-ROI processing, consider using a for loop.
    """
    # Store debug
    debug = params.debug
    params.debug = None

    if len(roi.contours) > 1:
        warn("received a multi-ROI but only the first ROI will be used. Consider using a for loop for multi-ROI")

    roi_contour = roi.contours[0]
    object_contour = obj
    obj_hierarchy = hierarchy

    # Create an empty grayscale (black) image the same dimensions as the input image
    mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
    cv2.drawContours(mask, object_contour, -1, (255), -1, lineType=8, hierarchy=obj_hierarchy)

    # Create a mask of the filled in ROI
    roi_mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
    roi_points = np.vstack(roi_contour[0])
    cv2.fillPoly(roi_mask, [roi_points], (255))

    # Allows user to find all objects that are completely inside or overlapping with ROI
    if roi_type.upper() in ('PARTIAL', 'LARGEST'):
        # Filter contours outside of the region of interest
        for c, _ in enumerate(object_contour):
            filtering_mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
            cv2.fillPoly(filtering_mask, [np.vstack(object_contour[c])], (255))
            overlap_img = _logical_operation(filtering_mask, roi_mask, 'and')
            # Delete contours that do not overlap at all with the ROI
            if np.sum(overlap_img) == 0:
                cv2.drawContours(mask, object_contour, c, (0), -1, lineType=8, hierarchy=obj_hierarchy)

        # Find the kept contours and area
        kept_cnt, kept_hierarchy = _cv2_findcontours(bin_img=mask)

        # Find the largest contour if roi_type is set to 'largest'
        if roi_type.upper() == 'LARGEST' and kept_cnt:
            index = np.argmax([cv2.contourArea(c) for c in kept_cnt])
            mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
            cv2.drawContours(mask, kept_cnt, contourIdx=index, color=(255), thickness=-1, hierarchy=kept_hierarchy, maxLevel=2)
            kept_cnt, kept_hierarchy = _cv2_findcontours(bin_img=mask)

    # Allows user to cut objects to the ROI (all objects completely outside ROI will not be kept)
    elif roi_type.upper() in ('CUTTO', 'WITHIN'):
        background1 = np.zeros(np.shape(img)[:2], dtype=np.uint8)
        background2 = np.zeros(np.shape(img)[:2], dtype=np.uint8)
        cv2.drawContours(background1, object_contour, -1, (255), -1, lineType=8, hierarchy=obj_hierarchy)
        roi_points = np.vstack(roi_contour[0])
        cv2.fillPoly(background2, [roi_points], (255))
        mask = cv2.multiply(background1, background2)
        kept_cnt, kept_hierarchy = _cv2_findcontours(bin_img=mask)

        # Filter out contours that touch the edge if roi_type is 'within'
        if roi_type.upper() == 'WITHIN' and kept_cnt:
            # make a mask with the outline of the ROI
            roi_outline_mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
            cv2.drawContours(image=roi_outline_mask, contours=roi_contour, contourIdx=-1,
                             color=255, thickness=1)
            # make empty mask to append to
            within_mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
            for c, _ in enumerate(kept_cnt):
                # for each contour make a mask with that contour filled
                filtering_mask = np.zeros(np.shape(img)[:2], dtype=np.uint8)
                cv2.fillPoly(filtering_mask, [np.vstack(kept_cnt[c])], (255))
                # check overlap with traced ROI
                overlap_img = _logical_operation(filtering_mask, roi_outline_mask, 'and')
                # check color in original mask, ie don't keep gaps that are 0s.
                # append contours fully within ROI to the within_mask
                if not overlap_img.any() and kept_hierarchy[0][c][3] == -1:
                    cv2.drawContours(within_mask, kept_cnt, c,
                                     int(img[kept_cnt[c][0][0][1], kept_cnt[c][0][0][0]]),
                                     -1, lineType=8, hierarchy=kept_hierarchy)
            mask = within_mask
            kept_cnt, kept_hierarchy = _cv2_findcontours(bin_img=mask)
    else:
        # Reset debug mode
        params.debug = debug
        fatal_error('ROI Type ' + str(roi_type) + ' is not "cutto", "largest", "within" or "partial"!')

    # Reset debug mode
    params.debug = debug

    return kept_cnt, kept_hierarchy, mask


def _iterate_analysis(img, labeled_mask, n_labels, label, function, **kwargs):
    """Iterate over labels and apply an analysis function.
    Inputs:
    img      = image to be used for visualization
    mask     = labeled mask
    n_labels = number of expected labels
    label    = label parameter, modifies the variable name of observations recorded
    function = analysis function to apply to each submask
    kwargs   = additional keyword arguments to pass to the analysis function

    :param img: np.ndarray
    :param mask: np.ndarray
    :param n_labels: int
    :param label: str
    :param function: function
    :param kwargs: dict
    """
    # Set labels to label
    labels = label
    # If label is a string, make a list of labels
    if isinstance(label, str):
        labels = [label] * n_labels
    # If the length of the labels list is not equal to the number of labels, raise an error
    if len(labels) != n_labels:
        fatal_error(f"Number of labels ({len(labels)}) does not match number of objects ({n_labels})")
    mask_copy = np.copy(labeled_mask)
    if len(np.unique(mask_copy)) == 2 and np.max(mask_copy) == 255:
        mask_copy = np.where(mask_copy == 255, 1, 0).astype(np.uint8)
    for i in range(1, n_labels + 1):
        submask = np.where(mask_copy == i, 255, 0).astype(np.uint8)
        img = function(img=img, mask=submask, label=f"{labels[i - 1]}_{i}", **kwargs)
    return img


def _object_composition(contours, hierarchy):
    """
    Groups objects into a single object, usually done after object filtering.

    Inputs:
    contours  = Contour tuple
    hierarchy = Contour hierarchy NumPy array

    Returns:
    group    = grouped contours list

    :param contours: tuple
    :param hierarchy: numpy.ndarray
    :return group: numpy.ndarray
    """
    stack = np.zeros((len(contours), 1))

    for c, _ in enumerate(contours):
        if hierarchy[0][c][2] == -1 and hierarchy[0][c][3] > -1:
            stack[c] = 0
        else:
            stack[c] = 1

    ids = np.where(stack == 1)[0]
    group = np.array([], dtype=np.int32)
    if len(ids) > 0:
        contour_list = [contours[i] for i in ids]
        group = np.vstack(contour_list)

    return group


def _grayscale_to_rgb(img):
    """
    Convert a grayscale image to an RGB image.

    Inputs:
    img = Grayscale or RGB image data

    Returns:
    img = RGB image data

    :param img: np.ndarray
    :return img: np.ndarray
    """
    if len(np.shape(img)) == 2:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)

    return img


def _rgb2lab(rgb_img, channel):
    """Convert image from RGB colorspace to LAB colorspace. Returns the specified subchannel as a gray image.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data
    channel : str
        color subchannel (l = lightness, a = green-magenta, b = blue-yellow)

    Returns
    -------
    numpy.ndarray
        grayscale image from one LAB color channel
    """
    # The allowable channel inputs are l, a or b
    channel = channel.lower()
    if channel not in ["l", "a", "b"]:
        fatal_error("Channel " + str(channel) + " is not l, a or b!")

    # Convert the input BGR image to LAB colorspace
    lab = cv2.cvtColor(rgb_img, cv2.COLOR_BGR2LAB)
    # Split LAB channels
    l, a, b = cv2.split(lab)
    # Create a channel dictionaries for lookups by a channel name index
    channels = {"l": l, "a": a, "b": b}
    return channels[channel]


def _rgb2hsv(rgb_img, channel):
    """Convert image from RGB colorspace to HSV colorspace. Returns the specified subchannel as a gray image.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data
    channel : str
        color subchannel (h = hue, s = saturation, v = value/intensity/brightness)

    Returns
    -------
    numpy.ndarray
        grayscale image from one HSV color channel
    """
    # The allowable channel inputs are h, s or v
    channel = channel.lower()
    if channel not in ["h", "s", "v"]:
        fatal_error("Channel " + str(channel) + " is not h, s or v!")

    # Convert the input BGR image to HSV colorspace
    hsv = cv2.cvtColor(rgb_img, cv2.COLOR_BGR2HSV)
    # Split HSV channels
    h, s, v = cv2.split(hsv)
    # Create a channel dictionaries for lookups by a channel name index
    channels = {"h": h, "s": s, "v": v}

    return channels[channel]


def _rgb2cmyk(rgb_img, channel):
    """Convert image from RGB colorspace to CMYK colorspace. Returns the specified subchannel as a gray image.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data
    channel : str
        color subchannel (c = cyan, m = magenta, y = yellow, k=black)

    Returns
    -------
    numpy.ndarray
        grayscale image from one CMYK color channel
    """
    # Set NumPy to ignore divide by zero errors
    _ = np.seterr(divide='ignore', invalid='ignore')
    # The allowable channel inputs are c, m , y or k
    channel = channel.lower()
    if channel not in ["c", "m", "y", "k"]:
        fatal_error("Channel " + str(channel) + " is not c, m, y or k!")

    # Create float
    bgr = rgb_img.astype(float)/255.

    # K channel
    k = 1 - np.max(bgr, axis=2)

    # C Channel
    c = (1 - bgr[..., 2] - k) / (1 - k)

    # M Channel
    m = (1 - bgr[..., 1] - k) / (1 - k)

    # Y Channel
    y = (1 - bgr[..., 0] - k) / (1 - k)

    # Convert the input BGR image to LAB colorspace
    cmyk = (np.dstack((c, m, y, k)) * 255).astype(np.uint8)
    # Split CMYK channels
    y, m, c, k = cv2.split(cmyk)
    # Create a channel dictionaries for lookups by a channel name index
    channels = {"c": c, "m": m, "y": y, "k": k}

    return channels[channel]


def _rgb2gray(rgb_img):
    """Convert image from RGB colorspace to Gray.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data

    Returns
    -------
    numpy.ndarray
        grayscale image
    """
    gray = cv2.cvtColor(rgb_img, cv2.COLOR_BGR2GRAY)

    return gray


def _logical_operation(bin_img1, bin_img2, operation):
    """Perform a logical operation on two binary images.

    Parameters
    ----------
    bin_img1 : numpy.ndarray
        First binary image
    bin_img2 : numpy.ndarray
        Second binary image
    operation : str
        Logical operation to perform ('and', 'or', or 'xor')

    Returns
    -------
    numpy.ndarray
        Resulting binary image after the logical operation
    """
    # Dictionary of operations
    operations = {
        'and': cv2.bitwise_and,
        'or': cv2.bitwise_or,
        'xor': cv2.bitwise_xor
    }
    # Check if the operation is valid
    if operation.lower() not in operations:
        fatal_error(f"Operation '{operation}' is not supported. Use 'and', 'or', or 'xor'.")
    # Perform the logical operation
    mask = operations[operation.lower()](bin_img1, bin_img2)
    return mask


def _scale_size(value, trait_type="linear"):
    """Convert size measurements to known scale parameter

    Parameters
    ----------
    value : float, list
        unscaled size value(s)
    trait_type : str
        type of size measurement, either "linear" (default) or "area"

    Returns
    -------
    float, list
        scaled trait value(s)
    """
    # Set the linear conversion rate
    conversion_rate = params.px_width
    # Update conversion rate if trait is per unit ^ 2
    if trait_type == "area":
        conversion_rate = params.px_width * params.px_height
    # Simple multiplication for size scaling a single value
    if type(value) is not list:
        return value * conversion_rate
    # Multiplication with list comprehension for lists of values
    return [x*conversion_rate for x in value]


def _identity(x, **kwargs):
    """Identity function for use in _rect_filter
    This may be useful if there are several outputs from a function passed to _rect_filter
    which would otherwise be difficult to manage
    Parameters
    ----------
    x : any
      An object
    **kwargs
      Other keyword arguments, ignored.
    """
    return x


def _rect_filter(img, roi=None, function=None, **kwargs):
    """Subset a rectangular section of image to apply function to
    Parameters
    ----------
    img : numpy.ndarray
        An image
    roi : plantcv Objects class
        A rectangular ROI as returned by plantcv.roi.rectangle
    function : function
        analysis function to apply to each submask
    **kwargs
        Other keyword arguments to pass to the analysis function.
    Returns
    -------
    any
        Return value depends on the function that is called. If no function is called then this is a numpy.ndarray.
    """
    if roi is None:
        xstart = 0
        ystart = 0
        xend = np.shape(img)[1]
        yend = np.shape(img)[0]
    else:
        xstart = roi.contours[0][0][0][0][0].astype("int32")
        ystart = roi.contours[0][0][0][0][1].astype("int32")
        xend = roi.contours[0][0][2][0][0].astype("int32")
        yend = roi.contours[0][0][2][0][1].astype("int32")
    # slice image to subset rectangle
    sub_img = img[ystart:yend, xstart:xend]
    # apply function
    if function is None:
        function = _identity

    return function(sub_img, **kwargs)


def _rect_replace(img, sub_img, roi):
    """
    Parameters
    ----------
    img : numpy.ndarray
        Full sized image
    sub_img : numpy.ndarray
        output from _rect_filter
    roi : plantcv Objects class
        A rectangular ROI as returned by plantcv.roi.rectangle
    Returns
    -------
    numpy.ndarray
    """
    if roi is None:
        # if no ROI then no subsetting was done, just return sub_img
        return sub_img

    # if subsetting was done then get coordinates, slice into main image, and return
    xstart = roi.contours[0][0][0][0][0].astype("int32")
    ystart = roi.contours[0][0][0][0][1].astype("int32")
    xend = roi.contours[0][0][2][0][0].astype("int32")
    yend = roi.contours[0][0][2][0][1].astype("int32")
    full_img = np.copy(img)
    full_img[ystart:yend, xstart:xend] = sub_img
    return full_img



# Binary image -> Skeleton

import os
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from skimage import morphology as skmorph


def skeletonize(mask):
    """Reduces binary objects to 1 pixel wide representations (skeleton)

    Inputs:
    mask       = Binary image data

    Returns:
    skeleton   = skeleton image

    :param mask: numpy.ndarray
    :return skeleton: numpy.ndarray
    """
    # Convert mask to boolean image, rather than 0 and 255 for skimage to use it
    skeleton = skmorph.skeletonize(mask.astype(bool))

    skeleton = skeleton.astype(np.uint8) * 255

    _debug(visual=skeleton, filename=os.path.join(params.debug_outdir, f"{params.device}_skeleton.png"))

    return skeleton



#!/usr/bin/env python
import os
import sys
import argparse
import time
import datetime
import plantcv.parallel
import plantcv.utils
import tempfile
import shutil


# Parse command-line arguments
###########################################
def options():
    """Parse command line options.

    Args:

    Returns:
        argparse object.
    Raises:
        IOError: if dir does not exist.
        IOError: if workflow does not exist.
        IOError: if the metadata file SnapshotInfo.csv does not exist in dir when flat is False.
        ValueError: if adaptor is not phenofront or dbimportexport.
        ValueError: if a metadata field is not supported.
    """
    parser = argparse.ArgumentParser(description='Parallel imaging processing with PlantCV.')
    config_grp = parser.add_argument_group("CONFIG")
    config_grp.add_argument("--template", required=False, help="Create a template configuration file.")
    run_grp = parser.add_argument_group("RUN")
    run_grp.add_argument("--config", required=False,
                         help="Input configuration file (created using the --template option).")
    args = parser.parse_args()

    # Create a config
    config = plantcv.parallel.WorkflowConfig()

    # Create a template configuration file if requested
    if args.template:
        config.save_config(config_file=args.template)
        sys.exit()

    # Import a configuration if provided
    if args.config:
        config.import_config(config_file=args.config)

    if not config.validate_config():
        raise ValueError("Invalid configuration file. Check errors above.")
    return config
###########################################


# Run the main program
###########################################
def main():
    """Main program.

    Args:

    Returns:

    Raises:

    """
    # Get options
    config = options()

    # Job start time
    start_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    print("Starting run " + start_time + '\n', file=sys.stderr)

    # Create temporary directory for job
    if config.tmp_dir is not None:
        os.makedirs(config.tmp_dir, exist_ok=True)
    config.tmp_dir = tempfile.mkdtemp(prefix=start_time + '_', dir=config.tmp_dir)

    # Create img_outdir
    os.makedirs(config.img_outdir, exist_ok=True)

    # Remove JSON results file if append=False
    if not config.append and os.path.exists(config.json):
        os.remove(config.json)

    # Read image metadata
    ###########################################
    parser_start_time = time.time()
    print("Reading image metadata...", file=sys.stderr)
    meta, _ = plantcv.parallel.metadata_parser(config=config)
    parser_clock_time = time.time() - parser_start_time
    print(f"Reading image metadata took {parser_clock_time} seconds.", file=sys.stderr)
    ###########################################

    # Process images
    ###########################################
    # Job builder start time
    job_builder_start_time = time.time()
    print("Building job list... ", file=sys.stderr)
    jobs = plantcv.parallel.job_builder(meta=meta, config=config)
    job_builder_clock_time = time.time() - job_builder_start_time
    print(f"Building job list took {job_builder_clock_time} seconds.", file=sys.stderr)

    # Parallel image processing time
    multi_start_time = time.time()
    print("Processing images... ", file=sys.stderr)
    cluster_client = plantcv.parallel.create_dask_cluster(cluster=config.cluster, cluster_config=config.cluster_config)
    plantcv.parallel.multiprocess(jobs=jobs, client=cluster_client)
    multi_clock_time = time.time() - multi_start_time
    print(f"Processing images took {multi_clock_time} seconds.", file=sys.stderr)
    ###########################################

    # Compile image analysis results
    ###########################################
    # Process results start time
    process_results_start_time = time.time()
    print("Processing results... ", file=sys.stderr)
    plantcv.parallel.process_results(job_dir=config.tmp_dir, json_file=config.json)
    process_results_clock_time = time.time() - process_results_start_time
    print(f"Processing results took {process_results_clock_time} seconds.", file=sys.stderr)
    ###########################################

    # Convert json results to csv files
    ###########################################
    # Convert results start time
    convert_results_start_time = time.time()
    print("Converting json to csv... ", file=sys.stderr)
    plantcv.utils.json2csv(config.json, config.json)
    convert_results_clock_time = time.time() - convert_results_start_time
    print(f"Processing results took {convert_results_clock_time} seconds.", file=sys.stderr)
    ###########################################

    # Cleanup
    if config.cleanup is True:
        shutil.rmtree(config.tmp_dir)
###########################################



from plantcv.parallel.parsers import metadata_parser
from plantcv.parallel.inspect_dataset import inspect_dataset
from plantcv.parallel.job_builder import job_builder
from plantcv.parallel.process_results import process_results
from plantcv.parallel.multiprocess import multiprocess
from plantcv.parallel.multiprocess import create_dask_cluster
from plantcv.parallel.workflow_inputs import workflow_inputs, WorkflowInputs
from plantcv.parallel.workflowconfig import WorkflowConfig


__all__ = ["metadata_parser", "inspect_dataset", "job_builder", "process_results",
           "multiprocess", "create_dask_cluster", "WorkflowConfig", "workflow_inputs", "WorkflowInputs"]



"""Find leaf insertion angles with stem."""
import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv import fatal_error
from plantcv.plantcv import color_palette
from plantcv.plantcv.morphology.segment_tangent_angle import _slope_to_intersect_angle
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _find_tips, _iterative_prune, _logical_operation, _dilate, _closing


def segment_insertion_angle(skel_img, segmented_img, leaf_objects, stem_objects, size, label=None):
    """
    Find leaf insertion angles in degrees of skeleton segments.

    Fit a linear regression line to the stem. Use `size` pixels on the portion of leaf next to the stem to find a linear
    regression line, and calculate the angle between the two lines per leaf object.

    Parameters
    ----------
    skel_img : numpy.ndarray
        Skeletonized image.
    segmented_img : numpy.ndarray
        Segmented image to plot slope lines and intersection angles on.
    leaf_objects : list
        List of leaf segments.
    stem_objects : list
        List of stem segments.
    size : int
        Size of inner leaf used to calculate slope lines.
    label : str, optional
        Label parameter, modifies the variable name of observations recorded (default = pcv.params.sample_label).

    Returns
    -------
    labeled_img : numpy.ndarray
        Debugging image with angles labeled.
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    cols = segmented_img.shape[1]
    labeled_img = segmented_img.copy()
    segment_slopes = []
    insertion_segments = []
    insertion_hierarchies = []
    intersection_angles = []
    all_intersection_angles = []
    label_coord_x = []
    label_coord_y = []
    valid_segment = []
    pruned_away = []

    # Create a list of tip tuples to use for sorting
    tips, _, _ = _find_tips(skel_img)

    for i, cnt in enumerate(leaf_objects):
        # Draw leaf objects
        find_segment_tangents = np.zeros(segmented_img.shape[:2], np.uint8)
        cv2.drawContours(find_segment_tangents, leaf_objects, i, 255, 1, lineType=8)

        # Prune back ends of leaves
        pruned_segment = _iterative_prune(find_segment_tangents, size)

        # Segment ends are the portions pruned off
        segment_ends = find_segment_tangents - pruned_segment
        segment_end_obj, segment_end_hierarchy = _cv2_findcontours(bin_img=segment_ends)

        if not len(segment_end_obj) == 2:
            print("Size too large, contour with ID#", i, "got pruned away completely.")
            pruned_away.append(True)
        else:
            # The contour can have insertion angle calculated
            pruned_away.append(False)
            valid_segment.append(cnt)

            # Determine if a segment is leaf end or leaf insertion segment
            for j, obj in enumerate(segment_end_obj):

                segment_plot = np.zeros(segmented_img.shape[:2], np.uint8)
                cv2.drawContours(segment_plot, obj, -1, 255, 1, lineType=8)
                segment_plot = _dilate(segment_plot, 3, 1)
                overlap_img = _logical_operation(segment_plot, tips, "and")

                # If none of the tips are within a segment_end then it's an insertion segment
                if np.sum(overlap_img) == 0:
                    insertion_segments.append(segment_end_obj[j])
                    insertion_hierarchies.append(segment_end_hierarchy[0][j])

            # Store coordinates for labels
            label_coord_x.append(leaf_objects[i][0][0][0])
            label_coord_y.append(leaf_objects[i][0][0][1])

    # Create a color scale, use a previously stored scale if available
    rand_color = color_palette(num=len(valid_segment), saved=True)

    for i, cnt in enumerate(valid_segment):
        cv2.drawContours(labeled_img, valid_segment, i, rand_color[i], params.line_thickness, lineType=8)

    # Plot stem segments
    combined_stem = _combine_stem(segmented_img, stem_objects)
    # Find slope of the stem
    [vx, vy, x, y] = cv2.fitLine(combined_stem[0], cv2.DIST_L2, 0, 0.01, 0.01)
    stem_slope = -vy / vx
    stem_slope = stem_slope[0]
    lefty = int(np.array((-x * vy / vx) + y).item())
    righty = int(np.array(((cols - x) * vy / vx) + y).item())
    cv2.line(labeled_img, (cols - 1, righty), (0, lefty), (150, 150, 150), 3)

    for t, segment in enumerate(insertion_segments):
        # Find line fit to each segment
        [vx, vy, x, y] = cv2.fitLine(segment, cv2.DIST_L2, 0, 0.01, 0.01)
        slope = -vy / vx
        left_list = int(np.array((-x * vy / vx) + y).item())
        right_list = int(np.array(((cols - x) * vy / vx) + y).item())
        segment_slopes.append(slope[0])

        # Draw slope lines if possible
        if abs(slope) > 1000000:
            print("Slope of contour with ID#", t, "is", slope, "and cannot be plotted.")
        else:
            cv2.line(labeled_img, (cols - 1, right_list), (0, left_list), rand_color[t], 1)

        # Store intersection angles between insertion segment and stem line
        intersection_angle = _slope_to_intersect_angle(slope[0], stem_slope)
        # Function measures clockwise but we want the acute angle between stem and leaf insertion
        if intersection_angle > 90:
            intersection_angle = 180 - intersection_angle
        intersection_angles.append(intersection_angle)

    # Compile list of measurements where there is a 'NA' where pruned away segments would go.
    intersection_angles_editing = intersection_angles.copy()
    for j in pruned_away:
        if j:
            all_intersection_angles.append('NA')
        else:
            all_intersection_angles.append(intersection_angles_editing[0])
            intersection_angles_editing.remove(intersection_angles_editing[0])

    segment_ids = []

    for i, cnt in enumerate(insertion_segments):
        # Label slope lines
        w = label_coord_x[i]
        h = label_coord_y[i]
        text = f"{intersection_angles[i]:0,.2f}"
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)
        segment_ids.append(i)

    outputs.add_observation(sample=label, variable='segment_insertion_angle', trait='segment insertion angle',
                            method='plantcv.plantcv.morphology.segment_insertion_angle', scale='degrees', datatype=list,
                            value=all_intersection_angles, label=segment_ids)

    _debug(visual=labeled_img,
           filename=os.path.join(params.debug_outdir, f"{params.device}_segment_insertion_angles.png"))

    return labeled_img


def _combine_stem(segmented_img, stem_objects, maxiter=50):
    """
    Combine multiple stem contours into a single stem contour.

    Parameters
    ----------
    segmented_img : numpy.ndarray
        Segmented image used for plotting and processing.
    stem_objects : list of numpy.ndarray
        List of contours representing stem objects.
    maxiter : int, optional
        Maximum number of iterations to attempt merging stem objects (default is 50).

    combined_stem : list of numpy.ndarray
        List containing the merged stem contour.

    Returns
    -------
    combined_stem : list of numpy.ndarray
        List containing the merged stem contour.

    Raises
    ------
    RuntimeError
        If stem objects cannot be combined into a single contour within `maxiter` iterations.
    """
    stem_img = np.zeros(segmented_img.shape[:2], np.uint8)
    cv2.drawContours(stem_img, stem_objects, -1, 255, 2, lineType=8)
    stem_img = _closing(stem_img)
    combined_stem, _ = _cv2_findcontours(bin_img=stem_img)

    # Make sure stem objects are a single contour
    loop_count = 0
    while len(combined_stem) > 1 and loop_count < maxiter:
        loop_count += 1
        stem_img = _dilate(stem_img, 2, 1)
        stem_img = _closing(stem_img)
        combined_stem, _ = _cv2_findcontours(bin_img=stem_img)
    if len(combined_stem) > 1:
        fatal_error('Unable to combine stem objects.')
    return combined_stem



from plantcv.plantcv import fatal_error


def _show_dataarray(img, **kwargs):
    """
    Plot facetted images from xarray dataarray

    Inputs:
    img     - dataarray to display
    kwargs  - key-value arguments to xarray.plot method

    :param img: xr.core.dataarray.DataArray
    :param kwargs: dict of arguments recognized by xarray.plot.plot
    """
    # check for kwargs col and row. col and row are removed from kwargs!
    # the default for col and row are None
    col = kwargs.pop('col', None)
    row = kwargs.pop('row', None)

    col_or_row_given = col is not None or row is not None
    if not col_or_row_given:
        fatal_error('You need to specify `col` or `row` with which to facet the xarray images. '
                    'We only support xarray facetted image plots using pcolormesh() in pcv functions. '
                    'For other types of plots please use xarray plotting methods and matplotlib directly.')

    contains_xy = len([dim for dim in ['x', 'y'] if dim in img.dims]) == 2
    if not contains_xy:
        fatal_error('You are missing x and y dimensions. '
                    'We only support xarray facetted image plots using pcolormesh() in pcv functions. '
                    'For other types of plots please use xarray plotting methods and matplotlib directly.')

    try:
        # need to force pcolormesh() for case when dim in col or row has length 1
        # https://github.com/pydata/xarray/issues/620
        fig_handle = img.plot.pcolormesh(col=col, row=row, **kwargs)
    except ValueError as err:
        raise ValueError(f'You are trying to plot shape {img.shape} but you should have exactly 2 dimensions in '
                         'addition those specified by `col` and `row`.') from err

    return fig_handle



# Rotate hyperspectral datacubes counterclockwise (in increments of 90 degrees)

# imports
import numpy as np
import os
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import Spectral_data


def rot90(spectral_data, k):
    """This function allows you rotate hyperspectral image data counterclockwise in 90 degree increments.

    Inputs:
    spectral_data   = Hyperspectral data instance
    k               = Number of times the array is rotated by 90 degrees

    Returns:
    rot_hsi         = Rotated array data

    :param spectral_data: __main__.Spectral_data
    :param k: int
    :return rot_hsi: __main__.Spectral_data
    """
    # Extract array data and rotate with numpy function
    rot_array = np.rot90(m=spectral_data.array_data, k=k)
    rot_rgb = np.rot90(m=spectral_data.pseudo_rgb, k=k)

    # Dimensions of datacube might change depending on degree of rotation
    dims = np.shape(rot_array)

    # Create new spectral data object with rotated image data and metadata
    rot_hsi = Spectral_data(array_data=rot_array, min_wavelength=spectral_data.min_wavelength,
                            max_wavelength=spectral_data.max_wavelength, min_value=spectral_data.min_value,
                            max_value=spectral_data.max_value, d_type=spectral_data.d_type,
                            wavelength_dict=spectral_data.wavelength_dict, samples=int(dims[2]), lines=int(dims[0]),
                            interleave=spectral_data.interleave, wavelength_units=spectral_data.wavelength_units,
                            array_type="datacube", pseudo_rgb=rot_rgb, default_bands=spectral_data.default_bands,
                            filename="rot_k" + str(k) + spectral_data.filename)

    _debug(visual=rot_rgb,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_pseudo_rgb_rot_k" + str(k) + ".png"))

    return rot_hsi



# Sobel filtering

import cv2
import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rect_filter, _rect_replace
from plantcv.plantcv import params


def sobel_filter(gray_img, dx, dy, ksize, roi=None):
    """
    This is a filtering method used to identify and highlight gradient edges/features using the 1st derivative.
    Typically used to identify gradients along the x-axis (dx = 1, dy = 0) and y-axis (dx = 0, dy = 1) independently.
    Performance is quite similar to Scharr filter. Used to detect edges / changes in pixel intensity. ddepth = -1
    specifies that the dimensions of output image will be the same as the input image.

    Inputs:
    gray_img = Grayscale image data
    dx       = derivative of x to analyze
    dy       = derivative of x to analyze
    ksize    = specifies the size of the kernel (must be an odd integer: 1,3,5, ... , 31)
    roi      = optional rectangular roi to apply sobel filter within

    Returns:
    sb_img   = Sobel filtered image

    :param gray_img: numpy.ndarray
    :param dx: int
    :param dy: int
    :param ksize: int
    :param roi: plantcv.plantcv.Objects
    :return sb_img: numpy.ndarray
    """
    sub_sb_img = _rect_filter(gray_img, roi, function=cv2.Sobel,
                              **{"ddepth": -1, "dx": dx, "dy": dy, "ksize": ksize})
    sb_img = _rect_replace(gray_img, sub_sb_img, roi)

    fname = str(params.device) + '_sb_img_dx' + str(dx) + '_dy' + str(dy) + '_kernel' + str(ksize) + '.png'
    _debug(visual=sb_img,
           filename=os.path.join(params.debug_outdir, fname),
           cmap='gray')

    return sb_img



# Pseudocolor any grayscale image

import os
import numpy as np
from cv2 import cvtColor, COLOR_BGR2RGB
from matplotlib import pyplot as plt
from plantcv.plantcv import params
from plantcv.plantcv import fatal_error
from plantcv.plantcv.apply_mask import apply_mask


def pseudocolor(gray_img, mask=None, cmap=None, background="image", min_value=0, max_value=255,
                axes=True, colorbar=True, title=None, bad_mask=None, bad_color="red"):
    """Pseudocolor any grayscale image to custom colormap

    Inputs:
    gray_img    = grayscale image data
    mask        = (optional) binary or labeled mask
    cmap        = (optional) colormap. default is the matplotlib default, viridis
    background  = (optional) background color/type, options are "image", "white", "black", or an array object (image)
                  (requires a mask). default = 'image'
    min_value   = (optional) minimum value for range of interest. default = 0
    max_value   = (optional) maximum value for range of interest. default = 255
    axes        = (optional) if False then x- and y-axis won't be displayed, nor will the title. default = True
    colorbar    = (optional) if False then colorbar won't be displayed. default = True
    title       = (optional) custom title for the plot gets drawn if title is not None. default = None
    bad_mask    = (optional) binary mask of pixels with "bad" values, e.g. nan or inf or any other values considered
                  to be not informative and to be excluded from analysis. default = None
    bad_color   = (optional) desired color to show "bad" pixels. default = "red"
    Returns:
    pseudo_image = pseudocolored image

    :param gray_img: numpy.ndarray
    :param mask: numpy.ndarray
    :param cmap: str
    :param background: str, numpy.ndarray
    :param min_value: numeric
    :param max_value: numeric
    :param axes: bool
    :param colorbar: bool
    :param title: str
    :param bad_mask: numpy.ndarray
    :param bad_color: str
    :return pseudo_image: matplotlib.figure.Figure
    """
    # Auto-increment the device counter
    params.device += 1

    # Make copies of the gray image
    gray_img1 = np.copy(gray_img)

    # Check if the image is grayscale
    if len(np.shape(gray_img)) != 2:
        fatal_error("Image must be grayscale.")

    bad_idx, bad_idy = [], []

    # Create a whole-image mask if mask is None
    if mask is None:
        mask = np.ones(np.shape(gray_img), dtype=np.uint8)

    # Convert the input mask to binary
    mask = mask.copy()
    # Set all values greater than 0 to 255
    mask[np.where(mask > 0)] = 255
    # Change type to unsigned 8-bit integer
    mask = mask.astype(np.uint8)

    # Apply the mask
    masked_img = np.ma.array(gray_img1, mask=~mask.astype(bool))

    # Set the background color or type
    bkgd = {
        "BLACK": {
            "image": np.zeros(np.shape(gray_img1), dtype=np.uint8),
            "cmap": "gray"
            },
        "WHITE": {
            "image": np.zeros(np.shape(gray_img1), dtype=np.uint8) + 255,
            "cmap": "gray_r"
            },
        "IMAGE": {
            "image": gray_img1,
            "cmap": "gray"
            }
    }

    # Check background types and proceed accordingly
    if isinstance(background, str) and background.upper() in ["IMAGE", "WHITE", "BLACK"]:
        bkg_img = bkgd[background.upper()]["image"]
        bkg_cmap = bkgd[background.upper()]["cmap"]
    elif isinstance(background, np.ndarray):
        bkg_img = np.copy(background)
        if bkg_img.shape[:2] != gray_img1.shape:
            fatal_error(f"The dimensions of the background image ({bkg_img.shape}) and "
                        f"the grayscale image ({gray_img1.shape}) are not the same.")
        # converting the array object to the correct color sequence
        bkg_img = cvtColor(bkg_img, COLOR_BGR2RGB)
        bkg_cmap = None
    else:
        fatal_error(f"Background type {background} is not supported. Please use 'white', 'black', 'image' or an array object.")

    if bad_mask is not None:
        debug_mode = params.debug
        params.debug = None
        bad_mask = apply_mask(bad_mask, mask, mask_color='black')
        bad_idx, bad_idy = np.where(bad_mask > 0)
        params.debug = debug_mode

    plt.figure()
    # Pseudocolor the image, plot the background first
    plt.imshow(bkg_img, cmap=bkg_cmap)
    # Overlay the masked grayscale image with the user input colormap
    plt.imshow(masked_img, cmap=cmap, vmin=min_value, vmax=max_value)
    plt.plot(bad_idy, bad_idx, '.', color=bad_color)

    if colorbar:
        plt.colorbar(fraction=0.033, pad=0.04)

    if axes and title is not None:
        # Include image title
        plt.title(title)
    else:
        # Remove axes
        plt.xticks([])
        plt.yticks([])

    # Store the current figure
    pseudo_img = plt.gcf()

    # Print or plot if debug is turned on
    if params.debug is not None:
        if params.debug == 'print':
            plt.savefig(os.path.join(params.debug_outdir, str(
                params.device) + '_pseudocolored.png'), dpi=params.dpi)
            plt.close()
        elif params.debug == 'plot':
            # Use non-blocking mode in case the function is run more than once
            plt.show(block=False)
    else:
        plt.close()

    return pseudo_img



# For classifying an image based on a trained kmeans clustering model output from train_kmeans.py

import os
import numpy as np
from joblib import load
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params
from plantcv.learn.train_kmeans import patch_extract
from plantcv.plantcv._helpers import _logical_operation


def predict_kmeans(img, model_path="./kmeansout.fit", patch_size=10):
    """Uses a trained, patch-based kmeans clustering model to predict clusters from an input image.

    Parameters
    ----------
    img : numpy.ndarray
        An image on which to predict clusters
    model_path : str, optional
        Path to directory where the trained model output is stored, by default "./kmeansout.fit"
    patch_size : int, optional
        Size of the NxN neighborhood around each pixel, by default 10

    Returns
    -------
    numpy.ndarray
        An labeled mask with the predicted clusters
    """
    kmeans = load(model_path)
    train_img = img.copy()

    before = after = int((patch_size - 1)/2)   # odd
    if patch_size % 2 == 0:   # even
        before = int((patch_size-2)/2)
        after = int(patch_size/2)

    # Padding
    if len(train_img.shape) == 2:  # gray
        train_img = np.pad(train_img, pad_width=((before, after), (before, after)), mode="edge")
    elif len(train_img.shape) == 3 and train_img.shape[2] == 3:  # rgb
        train_img = np.pad(train_img, pad_width=((before, after), (before, after), (0, 0)), mode="edge")

    # Shapes
    mg = np.floor(patch_size / 2).astype(np.int32)
    if len(train_img.shape) == 2:
        h, w = train_img.shape
    elif len(train_img.shape) == 3 and train_img.shape[2] == 3:
        h, w, _ = train_img.shape

    # Do the prediction
    train_patches = patch_extract(train_img, patch_size=patch_size)
    train_labels = kmeans.predict(train_patches)
    reshape_params = [[h - 2*mg + 1, w - 2*mg + 1], [h - 2*mg, w - 2*mg]]
    # Takes care of even vs odd numbered patch size reshaping
    labeled = train_labels.reshape(reshape_params[patch_size % 2][0], reshape_params[patch_size % 2][1])
    labeled = labeled.astype("uint8")
    _debug(visual=labeled, filename=os.path.join(params.debug_outdir, "_labeled_img.png"))
    return labeled


def mask_kmeans(labeled_img, k, cat_list=None):
    """Uses the predicted clusters from a target image to generate a binary mask.

    Parameters
    ----------
    labeled_img : numpy.ndarray
        An labeled mask with the predicted clusters
    k : int
        The number of clusters
    cat_list : list, optional
        A list of the numeric classes for composing a combined mask, by default None
        If empty, function prints all classes as separate masks

    Returns
    -------
    numpy.ndarray
        A binary mask with the specified clusters
    """
    if cat_list is None:
        mask_dict = {}
        L = [*range(k)]
        for i in L:
            mask_light = np.where(labeled_img == i, 255, 0).astype("uint8")
            _debug(visual=mask_light, filename=os.path.join(params.debug_outdir, "_kmeans_mask_"+str(i)+".png"))
            mask_dict[str(i)] = mask_light
        return mask_dict
    # Store debug
    debug = params.debug
    # Change to None so that logical_or does not plot each stepwise addition
    params.debug = None
    for idx, i in enumerate(cat_list):
        if idx == 0:
            mask_light = np.where(labeled_img == i, 255, 0).astype("uint8")
        else:
            mask_light = _logical_operation(mask_light, np.where(labeled_img == i, 255, 0).astype("uint8"), "or")
    params.debug = debug
    _debug(visual=mask_light, filename=os.path.join(params.debug_outdir, "_kmeans_combined_mask.png"))
    return mask_light



# Variance texture filter


import os
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rect_filter, _rect_replace
from plantcv.plantcv import params
from scipy.ndimage import generic_filter


def stdev_filter(img, ksize, borders='nearest', roi=None):
    """
    Creates a binary image from a grayscale image using skimage texture calculation for thresholding.
    This function is quite slow.

    Inputs:
    gray_img       = Grayscale image data
    ksize          = Kernel size for texture measure calculation
    borders        = How the array borders are handled, either 'reflect',
                     'constant', 'nearest', 'mirror', or 'wrap'
    roi            = optional rectangular ROI to apply filter within

    Returns:
    output         = Standard deviation values image

    :param img: numpy.ndarray
    :param ksize: int
    :param borders: str
    :param roi: plantcv.plantcv.Objects
    :return output: numpy.ndarray
    """
    # Make an array the same size as the original image
    output = np.zeros(img.shape, dtype=img.dtype)
    # Take the pieces of the empty mask and image in the ROI
    sub_zeros = _rect_filter(output, roi)
    sub_img = _rect_filter(img, roi)
    # Apply the texture function over the subset image
    generic_filter(sub_img, np.std, size=ksize, output=sub_zeros, mode=borders)
    # re-insert the subset into the full size mask
    replaced = _rect_replace(img, sub_zeros, roi)

    _debug(visual=replaced,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_variance.png"))

    return replaced
